"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import numpy as np
from sympy import lambdify, Expr

from qrisp.misc import bin_rep
from qrisp.circuit.standard_operations import op_list
from qrisp.circuit import ControlledOperation, ClControlledOperation


# Function to convert qrisp quantum circuits to Qiskit quantum circuits
def convert_to_qiskit(qc, transpile=False):
    from qiskit import ClassicalRegister, QuantumCircuit, QuantumRegister
    from qiskit.circuit import Parameter

    if transpile:
        from qrisp.circuit.transpiler import transpile

        qc = transpile(qc)

    # Create circuit

    qiskit_qc = QuantumCircuit()

    # This dic gives the qiskit qubits/clbits when presented with their identifier
    bit_dic = {}

    # Add Qubits
    for i in range(len(qc.qubits)):
        qiskit_qc.add_register(QuantumRegister(1, name=qc.qubits[i].identifier))

        bit_dic[qc.qubits[i].identifier] = qiskit_qc.qubits[-1]

    # Add Clbits
    for i in range(len(qc.clbits)):
        qiskit_qc.add_register(ClassicalRegister(1, name=qc.clbits[i].identifier))

        bit_dic[qc.clbits[i].identifier] = qiskit_qc.clbits[-1]

    symbol_param_dic = {}

    for i in range(len(qc.data)):
        op = qc.data[i].op

        params = list(op.params)

        qiskit_params = []

        for p in params:
            if isinstance(p, Expr):

                free_symbols = list(p.free_symbols)
                lambd_expr = lambdify(free_symbols, p)

                qiskit_symbs = []

                for s in free_symbols:
                    if not s in symbol_param_dic:
                        symbol_param_dic[s] = Parameter(str(s))

                    qiskit_symbs.append(symbol_param_dic[s])
                qiskit_params.append(lambd_expr(*qiskit_symbs))

            else:
                qiskit_params.append(p)

        params = qiskit_params

        # Prepare qiskit qubits
        qubit_list = [bit_dic[qubit.identifier] for qubit in qc.data[i].qubits]
        clbit_list = [bit_dic[clbit.identifier] for clbit in qc.data[i].clbits]

        if op.name in ["qb_alloc", "qb_dealloc"]:
            # if op.name == "qb_alloc":
            # continue
            continue
            temp = QuantumCircuit(1)
            qiskit_ins = temp.to_gate()
            qiskit_ins.name = op.name

        elif op.name == "gphase":
            temp_qc = QuantumCircuit(1)
            qiskit_ins = temp_qc.to_gate()
            qiskit_ins.name = "gphase"

        elif isinstance(op, ClControlledOperation):
            q_reg = QuantumRegister(op.num_qubits)

            # In Qiskit 1.3, the c_if interface was deprecated
            try:
                from qiskit.circuit import IfElseOp
    
                qregs = [qiskit_qc.qregs[qc.qubits.index(qb)] for qb in qc.data[i].qubits]
                body_qc = QuantumCircuit(*qregs)
                if op.base_op.definition:
                    body_qc = body_qc.compose(op.base_op.definition.to_qiskit())
                else:
                    base_qiskit_ins = create_qiskit_instruction(op.base_op, params)
                    body_qc.append(base_qiskit_ins, qubit_list)
                qiskit_ins = IfElseOp((clbit_list[0], 1), true_body=body_qc)
                clbit_list = []
            except ImportError:
                base_qiskit_ins = create_qiskit_instruction(op.base_op, params)
                cl_reg = ClassicalRegister(op.num_clbits)
                temp_qc = QuantumCircuit(q_reg, cl_reg)
                qiskit_ins = base_qiskit_ins.c_if(cl_reg, int(op.ctrl_state[::-1], 2))
                temp_qc.append(qiskit_ins, q_reg)
                qiskit_ins = temp_qc.to_instruction()
                qiskit_ins.name = op.name

        elif issubclass(op.__class__, ControlledOperation):
            base_name = op.base_operation.name

            if len(base_name) == 1:
                base_name = base_name.upper()

            if op.base_operation.definition:
                qiskit_definition = convert_to_qiskit(op.base_operation.definition)
                base_gate = qiskit_definition.to_gate()
                base_gate.name = op.base_operation.name
                qiskit_ins = base_gate.control(
                    len(op.controls), ctrl_state=op.ctrl_state[::-1]
                )

            else:

                if op.base_operation.name == "gphase":
                    qiskit_ins = create_qiskit_instruction(op, params)
                elif op.num_qubits == op.base_operation.num_qubits:
                    qiskit_ins = create_qiskit_instruction(op.base_operation, params)
                else:
                    base_gate = create_qiskit_instruction(op.base_operation, params)
                    qiskit_ins = base_gate.control(
                        len(op.controls), ctrl_state=op.ctrl_state[::-1]
                    )
        else:
            qiskit_ins = create_qiskit_instruction(op, params)

        # Append to qiskit circuit
        qiskit_qc.append(qiskit_ins, qubit_list, clbit_list)
    # Return result
    return qiskit_qc


def create_qiskit_instruction(op, params=[]):
    import qiskit.circuit.library.standard_gates as qsk_gates
    from qiskit.circuit import Measure, Reset, Barrier
    from qrisp.circuit import ControlledOperation
    from qiskit import QiskitError

    if op.name == "cx":
        if hasattr(op, "ctrl_state"):
            qiskit_ins = qsk_gates.XGate().control(ctrl_state=op.ctrl_state)
        else:
            qiskit_ins = qsk_gates.XGate().control()
    elif op.name == "cz":
        if hasattr(op, "ctrl_state"):
            qiskit_ins = qsk_gates.ZGate().control(ctrl_state=op.ctrl_state)
        else:
            qiskit_ins = qsk_gates.ZGate().control()
    elif op.name == "cy":
        if hasattr(op, "ctrl_state"):
            qiskit_ins = qsk_gates.YGate().control(ctrl_state=op.ctrl_state)
        else:
            qiskit_ins = qsk_gates.YGate().control()

    elif issubclass(op.__class__, ControlledOperation) and "gphase" not in op.name:
        base_name = op.base_operation.name

        if len(base_name) == 1:
            base_name = base_name.upper()

        if op.base_operation.definition:
            qiskit_definition = convert_to_qiskit(op.base_operation.definition)
            base_gate = qiskit_definition.to_gate()
            base_gate.name = op.base_operation.name
            qiskit_ins = base_gate.control(
                len(op.controls), ctrl_state=op.ctrl_state[::-1]
            )

        else:
            base_gate = create_qiskit_instruction(op.base_operation, params)
            qiskit_ins = base_gate.control(
                len(op.controls), ctrl_state=op.ctrl_state[::-1]
            )
    elif op.name == "rxx":
        qiskit_ins = qsk_gates.RXXGate(*params)
    elif op.name == "rzz":
        qiskit_ins = qsk_gates.RZZGate(*params)
    elif op.name == "ryy":
        qiskit_ins = qsk_gates.RYYGate(*params)
    elif op.name == "swap":
        qiskit_ins = qsk_gates.SwapGate()
    elif op.name == "measure":
        qiskit_ins = Measure()
    elif op.name == "barrier":
        qiskit_ins = Barrier(op.num_qubits)
    elif op.name == "cp":
        qiskit_ins = qsk_gates.CPhaseGate(params[0])
    elif op.name == "x":
        qiskit_ins = qsk_gates.XGate()
    elif op.name == "y":
        qiskit_ins = qsk_gates.YGate()
    elif op.name == "z":
        qiskit_ins = qsk_gates.ZGate()
    elif op.name == "rx":
        qiskit_ins = qsk_gates.RXGate(params[0])
    elif op.name == "ry":
        qiskit_ins = qsk_gates.RYGate(params[0])
    elif op.name == "rz":
        qiskit_ins = qsk_gates.RZGate(params[0])
    elif op.name == "sx":
        qiskit_ins = qsk_gates.SXGate()
    elif op.name == "h":
        qiskit_ins = qsk_gates.HGate()

    elif op.name == "p":
        qiskit_ins = qsk_gates.PhaseGate(params[0])
    elif op.name == "u1":
        qiskit_ins = qsk_gates.PhaseGate(params[0])
    elif op.name == "s":
        qiskit_ins = qsk_gates.SGate()
    elif op.name == "s_dg":
        qiskit_ins = qsk_gates.SGate().inverse()
    elif op.name == "sx":
        qiskit_ins = qsk_gates.SXGate()
    elif op.name == "sx_dg":
        qiskit_ins = qsk_gates.SXdgGate()
    elif op.name == "rzz":
        qiskit_ins = qsk_gates.RZZGate(params[0])
    elif op.name == "xxyy":
        qiskit_ins = qsk_gates.XXPlusYYGate(*params, label="(XX+YY)")
    elif op.name == "t":
        qiskit_ins = qsk_gates.TGate()
    elif op.name == "t_dg":
        qiskit_ins = qsk_gates.TGate().inverse()
    elif op.name == "u3":
        qiskit_ins = qsk_gates.U3Gate(*params)
    elif op.name == "r":
        qiskit_ins = qsk_gates.RGate(*params)
    elif op.name == "id":
        qiskit_ins = qsk_gates.IGate()
    elif op.name == "reset":
        qiskit_ins = Reset()
    elif op.definition:
        qiskit_definition = convert_to_qiskit(op.definition)
        try:
            qiskit_ins = qiskit_definition.to_gate()
        except QiskitError:
            qiskit_ins = qiskit_definition.to_instruction()

        qiskit_ins.name = op.name
    else:
        raise Exception("Could not convert operation " + str(op.name) + " to Qiskit")

    return qiskit_ins


op_dic = {op().name: op for op in op_list}
op_dic["u"] = op_dic["u3"]


def convert_from_qiskit(qiskit_qc):
    from qiskit.circuit import ControlledGate, ParameterExpression
    from qiskit import QuantumCircuit as QiskitQuantumCircuit

    from qrisp import Clbit, ControlledOperation, QuantumCircuit, Barrier, Qubit

    qc = QuantumCircuit()

    for i in range(qiskit_qc.num_qubits):
        q_reg = qiskit_qc.qubits[i]._register

        if q_reg is not None:

            if hasattr(q_reg, "_bits"):
                qb_list = q_reg._bits
            else:
                qb_list = list(q_reg)

            qubit_name = q_reg.name
            if q_reg.size > 1:
                qubit_name += "." + str(qb_list.index(qiskit_qc.qubits[i]))

            qc.add_qubit(Qubit(qubit_name))
        else:
            qc.add_qubit()

    for i in range(qiskit_qc.num_clbits):
        cl_reg = qiskit_qc.clbits[i]._register

        clbit_name = cl_reg.name
        if cl_reg.size > 1:
            clbit_name += "." + str(cl_reg._bits.index(qiskit_qc.clbits[i]))

        qc.add_clbit(Clbit(clbit_name))

    qb_dic = {qiskit_qc.qubits[i]: qc.qubits[i] for i in range(qiskit_qc.num_qubits)}
    cb_dic = {qiskit_qc.clbits[i]: qc.clbits[i] for i in range(qiskit_qc.num_clbits)}

    from sympy import sympify

    for i in range(len(qiskit_qc.data)):
        qiskit_op = qiskit_qc.data[i][0]

        if hasattr(qiskit_op, "condition_bits"):
            condition_bits = [cb_dic[cb] for cb in qiskit_op.condition_bits]
            if len(condition_bits):
                condition_value = qiskit_op.condition[1]
        elif hasattr(qiskit_op, "is_control_flow") and qiskit_op.is_control_flow():
            condition_bits = [cb_dic[cb] for cb in qiskit_qc.data[i].clbits]
            condition_value = qiskit_op.condition[1]
        else:
            condition_bits = []

        params = list(qiskit_op.params)

        qrisp_params = []

        while len(params):
            p = params.pop(0)
            if isinstance(p, np.ndarray):
                params = list(p.flatten()) + params
            elif isinstance(p, np.number):
                qrisp_params.append(p.item())
            elif isinstance(p, ParameterExpression):

                lambd_expr = sympify(ParameterExpression.sympify(p))

                qrisp_params.append(lambd_expr)

            elif isinstance(p, (float, int)):
                qrisp_params.append(float(p))
            elif isinstance(p, complex):
                qrisp_params.append(p)
            elif isinstance(p, QiskitQuantumCircuit):
                qiskit_op = p.to_gate()
                break
            else:
                print(p)
                raise Exception(f"Could not convert parameter type {type(p)}")

        params = qrisp_params

        if isinstance(qiskit_op, ControlledGate):
            controlled_gate = True
            qiskit_op = qiskit_op.base_gate
        else:
            controlled_gate = False

        try:
            op = op_dic[qiskit_op.name](*params)
        except KeyError:
            try:
                op = op_dic[qiskit_op.name.lower()](*params)
            except KeyError:
                if qiskit_op.definition is not None:
                    op = convert_from_qiskit(qiskit_op.definition).to_gate(
                        name=qiskit_op.name
                    )
                else:
                    raise Exception(
                        "Could not convert Qiskit operation "
                        + str(qiskit_op.name)
                        + " to Qrisp"
                    )

        if controlled_gate:
            qiskit_op = qiskit_qc.data[i][0]
            op = ControlledOperation(
                base_operation=op,
                num_ctrl_qubits=qiskit_op.num_ctrl_qubits,
                ctrl_state=bin_rep(qiskit_op.ctrl_state, qiskit_op.num_ctrl_qubits)[
                    ::-1
                ],
            )

        if op.name == "barrier":
            op = Barrier(qiskit_op.num_qubits)

        qubits = [qb_dic[qb] for qb in qiskit_qc.data[i][1]]
        clbits = [cb_dic[cb] for cb in qiskit_qc.data[i][2]]

        if condition_bits:
            clbits += condition_bits
            op = op.c_if(len(condition_bits), condition_value)

        qc.append(op, qubits, clbits)

    return qc



"""
********************************************************************************
* Copyright (c) 2024 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""


def test_IQAE_integration():
    from qrisp import QuantumFloat, QuantumBool, control, z, h, ry, IQAE
    import numpy as np

    # We compute the integral of f(x)=(sin(x))^2 from 0 to 1
    def state_function(inp, tar):
        h(inp) # Distribution
    
        N = 2**inp.size
        for k in range(inp.size):
            with control(inp[k]):
                ry(2**(k+1)/N,tar)

    n = 6 # 2^n sampling points for integration
    inp = QuantumFloat(n,-n)
    tar = QuantumBool()
    input_list = [inp, tar]

    eps = 0.01
    alpha = 0.01

    a = IQAE(input_list, state_function, eps=eps, alpha=alpha)
    assert np.abs(a-0.26716231971793425)<0.01





"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""


from qrisp import h, QuantumFloat, multi_measurement, auto_uncompute, QuantumBool, mcx, cx
from qrisp.quantum_backtracking import QuantumBacktrackingTree
import numpy as np
from sympy import nsimplify, Float

def test_backtracking():
    
    # This test is depracated (i.e., for previous implementation of backtracking). 
    # In this example [1,1,1] is accepted and rejected and can be reached in the backtracking tree!
    """
    @auto_uncompute    
    def P(tree):
        temp_0 = (tree.h == 0)
        predicate = QuantumBool()
        mcx(list(tree.branch_qa) + list(temp_0), predicate)
        
        return predicate

    @auto_uncompute    
    def Q(tree):
        return (tree.h != tree.max_depth) & (tree.h != (tree.max_depth-1)) & (tree.branch_qa[0] == tree.branch_qa[1])

    tree = QuantumBacktrackingTree(3, QuantumFloat(1, name = "branch_qf*"), P, Q)
    tree.init_node([])
    res = tree.estimate_phase(4)
    
    mes_res = res.get_measurement()
    
    assert mes_res[0] < 0.25
    """

    # Reject states with two consecutive 0's or 1's in path
    @auto_uncompute
    def reject(tree):

        reject_qbl = QuantumBool()
        for i in range(tree.max_depth-1):
            mcx([tree.h[i],tree.branch_qa[i],tree.branch_qa[i+1]],reject_qbl,ctrl_state="100")
            mcx([tree.h[i],tree.branch_qa[i],tree.branch_qa[i+1]],reject_qbl)
        return reject_qbl
    
    @auto_uncompute
    def accept(tree):

        height_condition = (tree.h == 0)

        path_condition = QuantumBool()
        mcx(tree.branch_qa[::-1], path_condition, ctrl_state = "001")

        return height_condition & path_condition

    tree = QuantumBacktrackingTree(max_depth = 3, branch_qv = QuantumFloat(1), accept = accept, reject = reject)
    tree.init_node([])
    qpe_res = tree.estimate_phase(precision = 4)

    mes_res = qpe_res.get_measurement()

    assert mes_res[0] < 0.25
    
    
    @auto_uncompute    
    def P(tree):
        predicate = QuantumBool()
        mcx(list(tree.branch_qa), predicate)
        return predicate

    @auto_uncompute    
    def Q(tree):
        return QuantumBool()
    
    tree = QuantumBacktrackingTree(3, QuantumFloat(1, name = "branch_qf*"), P, Q)
    tree.init_node([])
    res = tree.estimate_phase(4)
    
    mes_res = res.get_measurement()
    
    assert mes_res[0] > 0.375
    
    for i in range(1,5):
    
        tree = QuantumBacktrackingTree(i, QuantumFloat(1, name = "branch_qf*"), P, Q)
        
        tree.init_phi(i*[1])
        
        temp_sv_0 = tree.qs.statevector().expand().evalf()
        tree.quantum_step()
        temp_sv_1 = tree.qs.statevector().expand().evalf()
        
        diff = (temp_sv_0 - temp_sv_1).expand().evalf()
        diff_atoms = list(diff.atoms(Float))
        s = sum(diff_atoms)
        assert abs(s) < 1E-4
        
        tree = QuantumBacktrackingTree(i, QuantumFloat(1, name = "branch_qf*"), P, Q)
        tree.init_node(i*[1])
        
        temp_sv_0 = tree.qs.statevector().expand().evalf()
        tree.qstep_diffuser(even = True)
        temp_sv_1 = tree.qs.statevector().expand().evalf()
        
        diff = (temp_sv_0 - temp_sv_1).expand().evalf()
        diff_atoms = list(diff.atoms(Float))
        s = sum(diff_atoms)

        assert abs(s) < 1E-4
    
        
    tree = QuantumBacktrackingTree(3, QuantumFloat(1, name = "branch_qf*"), P, Q)
    solution = tree.find_solution(precision = 4)
    assert solution == [1,1,1]
    
    
    #Test non-binary tree
    
    @auto_uncompute    
    def accept(tree):
        height_condition = tree.h == 0
        path_condition = QuantumBool()
        qubit_list = sum([list(qv) for qv in tree.branch_qa.flatten()], [])
        mcx(qubit_list, path_condition, method = "balauca")
        return height_condition & path_condition

    @auto_uncompute    
    def reject(tree):
        return QuantumBool()

    max_depth = 2
    
    tree = QuantumBacktrackingTree(max_depth, QuantumFloat(2), accept, reject, subspace_optimization = True)
    tree.init_node([])
    qpe_res = tree.estimate_phase(3)
    
    mes_res = qpe_res.get_measurement()
    
    assert mes_res[0] > 0.375
    
    @auto_uncompute    
    def reject(tree):
        return tree.h == 1

    max_depth = 2
    
    tree = QuantumBacktrackingTree(max_depth, QuantumFloat(2), accept, reject, subspace_optimization = True)
    tree.init_node([])
    qpe_res = tree.estimate_phase(3)
    
    mes_res = qpe_res.get_measurement()
    
    assert mes_res[0] < 0.25



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import numpy as np

from qrisp.core import cx, swap


def demux(
    input,
    ctrl_qv,
    output=None,
    ctrl_method=None,
    permit_mismatching_size=False,
    parallelize_qc=False,
):
    """
    This functions allows moving an input value into an iterable output, where the
    position is specified by a ``QuantumFloat``. Demux is short for demultiplexer and
    is a standard component in `classical electrical circuitry
    <https://en.wikipedia.org/wiki/Multiplexer>`_.

    Demux can either move qubit states into a QuantumVariable or ``QuantumVariables``
    into ``QuantumArrays``.

    This function can also be used to "in-place demux" the 0-th entry of an iterable to
    the position specified by ``ctrl_qv``. For more information on this, check the
    second example.

    Parameters
    ----------
    input : Qubit or QuantumVariable
        The input value that is supposed to be moved.
    ctrl_qv : QuantumFloat
        The QuantumFloat specifying to which output the input should be moved.
    output : QuantumVariable or QuantumArray, optional
        The output object, where the input should end up. By default, a new object
        (QuantumVariable or QuantumArray) is created. Note that when this parameter is
        given, it is guaranteed, that the 0-th entry will be moved to the desired
        position, the other entries can also be permuted away from their original
        position.
    ctrl_method : string, optional
        The ``ctrl_method`` string passed to the
        :ref:`control environment <ControlEnvironment>` to generate controlled swaps.
    permit_mismatching_size : bool, optional
        If set to False, an exception will be raised, if the state-space dimension of
        `ctrl_qv`` is differing from the amount of outputs. The default is False.
    parallelize_qc : bool, optional
        If set to True, this option reduces (de)allocates additional qubits to
        reduce the depth. The default is False.

    Raises
    ------
    Exception
        Tried to demux with mismatchingly sized control input.

    Returns
    -------
    output : QuantumVariable or QuantumArray
        The output object with the input signal placed at the index specified by
        ``ctrl_qv``.

    Examples
    --------

    We create a ``QuantumBool`` and demux it into a ``QuantumArray`` ::

        from qrisp import *

        qb = QuantumBool()
        qb.flip()

        index = QuantumFloat(2)

        h(index[1])

        res_array = demux(qb, index)

    >>> print(multi_measurement([index, res_array]))
    {(0, OutcomeArray([1., 0., 0., 0.])): 0.5, (2, OutcomeArray([0., 0., 1., 0.])): 0.5}

    Demux can also be used to move the 0-th entry of a ``QuantumArray`` in-place. ::

        qa = QuantumArray(shape = 4, qtype = qb)

        qa[0].flip()

        demux(qa[0], index, qa)

    >>> print(multi_measurement([index, qa]))
    {(0, OutcomeArray([1., 0., 0., 0.])): 0.5, (2, OutcomeArray([0., 0., 1., 0.])): 0.5}

    For low-level manipulations, demux can move information within ``QuantumVariables``.
    ::

        qf = QuantumVariable(4)

        qf[:] = "1000"

        demux(qf[0], index, qf)

    >>> print(multi_measurement([index, qf]))
    {(0, '1000'): 0.5, (2, '0010'): 0.5}
    """

    from qrisp import QuantumArray, QuantumVariable, Qubit, control, swap

    if output is None:
        if isinstance(input, QuantumVariable):
            output = QuantumArray(input, 2 ** len(ctrl_qv))
        elif isinstance(input, Qubit):
            output = QuantumVariable(2 ** len(ctrl_qv))
        else:
            raise Exception("Don't know how to handle input type " + str(type(input)))
    else:
        if isinstance(output, QuantumArray):
            for qv in output.flatten()[1:]:
                if qv.name == input.name:
                    raise Exception(
                        "Tried to in-place demux QuantumArray entry,"
                        "which is not a 0-th position"
                    )
        elif isinstance(output, QuantumVariable):
            for qb in output.reg[1:]:
                if qb.identifier == input.identifier:
                    raise Exception(
                        "Tried to in-place demux QuantumVariable entry,"
                        "which is not a 0-th position"
                    )

    n = int(np.ceil(np.log2(len(output))))
    N = 2**n

    if len(output) != 2 ** len(ctrl_qv) and not permit_mismatching_size:
        raise Exception("Tried to demux with mismatching sized control input")

    if hash(input) != hash(output[0]):
        swap(input, output[0])

    if not len(ctrl_qv):
        return output

    if len(output) > 2 ** (len(ctrl_qv) - 1):
        with control(ctrl_qv[-1], ctrl_method=ctrl_method):
            swap(output[0], output[N // 2])
    else:
        demux(
            output[0],
            ctrl_qv[:-1],
            output,
            ctrl_method=ctrl_method,
            permit_mismatching_size=permit_mismatching_size,
        )
        return output

    if n > 1:

        if parallelize_qc:
            demux_ancilla = QuantumVariable(len(ctrl_qv) - 1)
            cx(ctrl_qv[:-1], demux_ancilla)
            ctrl_qubits = list(demux_ancilla)
        else:
            ctrl_qubits = ctrl_qv[:-1]

        demux(
            output[0],
            ctrl_qubits,
            # ctrl_qv[:-1],
            output[: N // 2],
            ctrl_method=ctrl_method,
            permit_mismatching_size=permit_mismatching_size,
            parallelize_qc=parallelize_qc,
        )

        demux(
            output[N // 2],
            ctrl_qv[:-1],
            output[N // 2 :],
            ctrl_method=ctrl_method,
            permit_mismatching_size=permit_mismatching_size,
            parallelize_qc=parallelize_qc,
        )
        if parallelize_qc:
            cx(ctrl_qv[:-1], demux_ancilla)
            demux_ancilla.delete()

    return output


def q_indexing(q_array, index):
    from qrisp import invert

    with invert():
        demux(q_array[0], index, q_array, ctrl_method="gray_pt")

    res = q_array[0].duplicate(init=True)

    demux(q_array[0], index, q_array, ctrl_method="gray_pt")

    return res


def q_swap_into(q_array, index, qv):
    from qrisp import invert, swap

    with invert():
        demux(q_array[0], index, q_array, ctrl_method="gray_pt")

    swap(q_array[0], qv)

    demux(q_array[0], index, q_array, ctrl_method="gray_pt")


def cyclic_shift(iterable, shift_amount=1):
    r"""
    Performs a cyclic shift of the values of an iterable with logarithmic depth.
    The shifting amount can be specified.


    Parameters
    ----------
    iterable : list[Qubit] or list[QuantumVariable] or QuantumArray
        The iterable to be shifted.
    shift_amount : integer or QuantumFloat, optional
        The iterable will be shifted by that amount. The default is 1.

    Examples
    --------

    We create a QuantumArray, initiate a sequence of increments and perform a cyclic shift.

    >>> from qrisp import QuantumFloat, QuantumArray, cyclic_shift
    >>> import numpy as np
    >>> qa = QuantumArray(QuantumFloat(3), 8)
    >>> qa[:] = np.arange(8)
    >>> cyclic_shift(qa, shift_amount = 2)
    >>> print(qa)
    {OutcomeArray([6, 7, 0, 1, 2, 3, 4, 5]): 1.0}

    We do something similar to demonstrate the shift by quantum values.
    For this we initiate a :ref:`QuantumFloat` in the superposition of 0, 1 and -3.

    >>> shift_amount = QuantumFloat(3, signed = True)
    >>> shift_amount[:] = {0 : 3**-0.5, 1: 3**-0.5, -3 : 3**-0.5}
    >>> qa = QuantumArray(QuantumFloat(3), 8)
    >>> qa[:] = np.arange(8)
    >>> cyclic_shift(qa, shift_amount)
    >>> print(qa)
    {OutcomeArray([0, 1, 2, 3, 4, 5, 6, 7]): 0.3333, OutcomeArray([7, 0, 1, 2, 3, 4, 5, 6]): 0.3333, OutcomeArray([3, 4, 5, 6, 7, 0, 1, 2]): 0.3333}
    """

    from qrisp import QuantumFloat, control, QuantumBool, cx

    if isinstance(shift_amount, QuantumFloat):

        if shift_amount.mshape[0] < 0:
            raise Exception("Tried to quantum shift by non-integer QuantumFloat")

        if shift_amount.signed:
            with control(shift_amount.sign()):
                cyclic_shift(iterable, -(2 ** (shift_amount.mshape[1])))

        for i in range(*shift_amount.mshape):
            with control(shift_amount.significant(i)):
                cyclic_shift(iterable, 2**i)

        return

    N = len(iterable)
    n = int(np.floor(np.log2(N)))

    if N == 0 or not shift_amount % N:
        return
    if shift_amount < 0:
        return cyclic_shift(iterable[::-1], -shift_amount)

    if shift_amount != 1:

        perm = np.arange(N)
        perm = (perm - shift_amount) % (N)

        permute_iterable(iterable, perm)
        return

    singular_shift(iterable[: 2**n])
    singular_shift([iterable[0]] + list(iterable[2**n :]), use_saeedi=True)


def singular_shift(iterable, use_saeedi=False):

    N = len(iterable)

    if N in [0, 1]:
        return

    if use_saeedi:
        # Strategy from https://arxiv.org/abs/1304.7516
        # Seems to perform worse when shifting by a quantum float
        # But better when the shift length is not a power of 2
        for i in range(N // 2):
            if (-i) % N == i + 1 or i + 1 >= N:
                continue
            swap(iterable[-i], iterable[i + 1])

        for i in range(N // 2):
            if (-i) % N == i + 2 or i + 2 >= N:
                continue
            swap(iterable[-i], iterable[i + 2])

    else:
        correction_indices = []
        for i in range(len(iterable) // 2):
            swap_tuple = (2 * i, 2 * i + 1)
            swap(iterable[swap_tuple[0]], iterable[swap_tuple[1]])
            correction_indices.append(swap_tuple[0])

        singular_shift([iterable[i] for i in correction_indices])


def to_cycles(perm):
    pi = {i: perm[i] for i in range(len(perm))}
    cycles = []

    while pi:
        elem0 = next(iter(pi))  # arbitrary starting element
        this_elem = pi[elem0]
        next_item = pi[this_elem]

        cycle = []
        while True:
            cycle.append(this_elem)
            del pi[this_elem]
            this_elem = next_item
            if next_item in pi:
                next_item = pi[next_item]
            else:
                break

        cycles.append(cycle)

    return cycles


def permute_iterable(iterable, perm):
    """
    Applies an arbitrary permutation to an iterable with logarithmic depth.

    Parameters
    ----------
    iterable : QuantumArray or List[QuantumVariable] or QuantumVariable or List[Qubit]
        The iterable to perform the permutation on.
    perm : List[integer]
        A list specifying the permutation.

    Examples
    --------

    We create a QuantumArray containing increments and apply a specified permutation.

    >>> from qrisp import QuantumFloat, QuantumArray, permute_iterable
    >>> import numpy as np
    >>> qa = QuantumArray(QuantumFloat(3), 8)
    >>> qa[:] = np.arange(8)
    >>> permute_iterable(qa, perm = [1,0,3,7,5,2,6,4])
    >>> print(qa)
    {OutcomeArray([1, 0, 3, 7, 5, 2, 6, 4]): 1.0}
    >>> print(qa.qs)

    ::

        QuantumCircuit:
        --------------
          qa.0: ────────────X──────────────────────
                            │
          qa.1: ──────X─────┼──────────────────────
                      │     │
          qa.2: ──────┼──X──┼──────────────────────
                ┌───┐ │  │  │
        qa_1.0: ┤ X ├─┼──┼──X──────────────────────
                └───┘ │  │
        qa_1.1: ──────X──┼─────────────────────────
                         │
        qa_1.2: ─────────X─────────────────────────

        qa_2.0: ───────────────────────────X───────
                ┌───┐                      │
        qa_2.1: ┤ X ├──────────────────────┼──X────
                └───┘                      │  │
        qa_2.2: ───────────────────────────┼──┼──X─
                ┌───┐                      │  │  │
        qa_3.0: ┤ X ├──────────X───────────┼──┼──┼─
                ├───┤          │           │  │  │
        qa_3.1: ┤ X ├──────────┼──X────────┼──┼──┼─
                └───┘          │  │        │  │  │
        qa_3.2: ───────────────┼──┼──X─────┼──┼──┼─
                               │  │  │     │  │  │
        qa_4.0: ──────X────────┼──┼──┼─────┼──┼──┼─
                      │        │  │  │     │  │  │
        qa_4.1: ──────┼──X─────┼──┼──┼─────┼──┼──┼─
                ┌───┐ │  │     │  │  │     │  │  │
        qa_4.2: ┤ X ├─┼──┼──X──┼──┼──┼─────┼──┼──┼─
                ├───┤ │  │  │  │  │  │     │  │  │
        qa_5.0: ┤ X ├─X──┼──┼──┼──┼──┼──X──X──┼──┼─
                └───┘    │  │  │  │  │  │     │  │
        qa_5.1: ─────────X──┼──┼──┼──┼──┼──X──X──┼─
                ┌───┐       │  │  │  │  │  │     │
        qa_5.2: ┤ X ├───────X──┼──┼──┼──┼──┼──X──X─
                └───┘          │  │  │  │  │  │
        qa_6.0: ───────────────┼──┼──┼──┼──┼──┼────
                ┌───┐          │  │  │  │  │  │
        qa_6.1: ┤ X ├──────────┼──┼──┼──┼──┼──┼────
                ├───┤          │  │  │  │  │  │
        qa_6.2: ┤ X ├──────────┼──┼──┼──┼──┼──┼────
                ├───┤          │  │  │  │  │  │
        qa_7.0: ┤ X ├──────────X──┼──┼──X──┼──┼────
                ├───┤             │  │     │  │
        qa_7.1: ┤ X ├─────────────X──┼─────X──┼────
                ├───┤                │        │
        qa_7.2: ┤ X ├────────────────X────────X────
                └───┘
        Live QuantumVariables:
        ---------------------
        QuantumFloat qa
        QuantumFloat qa_1
        QuantumFloat qa_2
        QuantumFloat qa_3
        QuantumFloat qa_4
        QuantumFloat qa_5
        QuantumFloat qa_6
        QuantumFloat qa_7

    """

    from sympy.combinatorics import Permutation

    inv_perm = list(Permutation(perm) ** -1)

    cycles = to_cycles(inv_perm)

    for c in cycles:
        cyclic_shift([iterable[i] for i in c], 1)



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import numpy as np
import networkx as nx

from qrisp.circuit import (
    QuantumCircuit,
    Operation,
    Qubit,
    PTControlledOperation,
    ControlledOperation,
    transpile,
    Instruction,
    fast_append,
    RXGate,
    RYGate,
    RZGate,
    RZZGate,
    PGate,
    GPhaseGate,
    ClControlledOperation
)
from qrisp.misc import get_depth_dic, retarget_instructions
from qrisp.permeability import optimize_allocations, parallelize_qc, lightcone_reduction

# The purpose of this function is to dynamically (de)allocate qubits when they are
# needed or not needed anymore. The qompiler function knows when a qubit is ready to
# deallocate (ie. it is in |0> state) due to a gate called QubitDealloc. After some
# preprocessing steps, this function will iterate through the data and dynamically
# (de)allocate if it is required. Another feature of this function is that it replaces
# mcx gates where the method is set to "auto" with implementations that fit the
# situation i.e. it takes into consideration how many clean/dirty ancilla qubits are
# available and synthesizes and mcx implementation based on this information.

# The parameter workspace can be given to extend the amount of qubits of the resulting
# QuantumCircuit by that amount. This extension can be helpfull because it gives the
# compiler more options to pick when allocating a new qubit. Since the allocation is
# based on a depth-reduction heuristic, giving more workspace results in many case in
# less depth circuits.


# It is however not only depth that can be improved by an increased workspace. Due to
# more clean/dirty ancillae beeing available, in many cases it is also possible to
# generate more efficient mcx implementations, thus also reducing the gate count.
def qompiler(
    qs,
    workspace=0,
    disable_uncomputation=True,
    intended_measurements=[],
    cancel_qfts=True,
    compile_mcm=False,
    gate_speed=None,
    use_dirty_anc_for_mcx_recomp=True,
):
    if len(qs.data) == 0:
        return QuantumCircuit(0)

    if gate_speed is None:
        gate_speed = lambda x: 1

    if compile_mcm is True:
        compile_mcm = "gidney"
    elif compile_mcm is False:
        compile_mcm = None

    with fast_append():
        qc = qs.copy()

        if not disable_uncomputation:
            local_qvs = qs.get_local_qvs()

            from qrisp.permeability.unqomp import uncompute_qc

            for qv in local_qvs:
                try:
                    qc = uncompute_qc(qc, qv.reg)
                except:
                    print(f"Warning: Automatic uncomputation for {qv.name} failed")

        # We now process the circuit in 3 important steps:
        #
        #   1. We use measurement reduction, i.e. we remove all gates that are not
        #   directly relevant to perform the intended measurement
        #
        #   2. We parallelize the circuit using the parallelization algorithm
        #
        #   3. We run reorder the circuit another time such that (de)allocations
        #   are executed in an advantageous order. The topological ordering algorithm
        #   behind the allocation is "stable" in the sense that it preserves the
        #   previously induced order by the parallelization algorithm.

        # To achieve these steps we view the circuit from different levels of abstraction
        # This is achieved by dissolving composite gates until a certain condition is met.
        # Such conditions can be specified using the transpile_predicate keyword.
        # The two relevant levels of abstractions are:
        #
        #   1. The permeability level: Composite gates are dissolved if they have incomplete
        #   permeability information. In turn this implies that the resulting circuit contains
        #   only gates with complete permeability information (such as adders, general arithmetic
        #   or multi controlled X-gates). The resulting DAG can therefore use high-level
        #   features for reordering.

        def permeability_transpile_predicate(op):
            for v in op.permeability.values():
                if v is None:
                    return True
            return False

        #   2. Allocation level: This level of abstraction leaves only a selected set of low
        #   level gates alive. This comes with the advantage that the allocation algorithm
        #   can also look into reordering the inner structure of high-level functions to also
        #   get a good allocation strategy for this type of functions.

        from qrisp.alg_primitives.logic_synthesis import LogicSynthGate
        from qrisp.alg_primitives.mcx_algs import GidneyLogicalAND, JonesToffoli
        from qrisp.alg_primitives.arithmetic import QuasiRZZ

        def allocation_level_transpile_predicate(op):

            if isinstance(op, PTControlledOperation):

                if op.base_operation.name == "x":
                    return False
                if op.base_operation.name == "p" and op.num_qubits == 2:
                    return False

            if isinstance(
                op, (LogicSynthGate, GidneyLogicalAND, JonesToffoli, QuasiRZZ)
            ):
                return False

            if op.name == "rzz":
                return False

            if "QFT" == op.name[:3] or "equal" in op.name or "less_than" in op.name:
                return False

            return True

        # Transpile to the first level

        qc = transpile(qc, transpile_predicate=permeability_transpile_predicate)

        if intended_measurements and len(qc.clbits) == 0:
            # This function reorders the circuit such that the intended measurements can
            # be executed as early as possible additionally, any instructions that are
            # not needed for the intended measurements are removed
            try:
                qc = lightcone_reduction(qc, intended_measurements)
                # pass
            except Exception as e:
                if "Unitary of operation " not in str(e):
                    raise e

        # Reorder circuit for parallelization.
        # For more details check the implementation of this function
        qc = parallelize_qc(qc, depth_indicator=gate_speed)
        qc = cancel_inverses(qc)

        # We now reorder the transpiled circuit to achieve a good (de)allocation order.
        # Reordering is performed based on the DAG representation of Unqomp.
        # The advantage of this representation is that
        # it abstracts away non-trivial commutation relations between permeable gates.
        # The actual ordering is performed by performing a topological sort on this dag.
        # Contrary to unqomp, we don't use a modified version of Kahns algorithm here
        # (for more information check the topological sort function)

        # The reordering process aims to find an order that minimizes that maximum
        # amount of qubits that is needed. Losely speaking it tries to reorder the data,
        # that allocations are performed as late as possible and deallocations are
        # performed as early as possible.

        # Note that the more aggressive we transpile, the more optimization is possible
        # through reordering i.e. less qubits. By reordering we however also destroy any
        # previous order which might have been intentionally picked to optimize depth.
        # In summary: Transpiling more aggressively leads to less qubits but more depth.

        transpiled_qc = transpile(
            qc, transpile_predicate=allocation_level_transpile_predicate
        )
        
        

        reordered_qc = optimize_allocations(transpiled_qc)

        if cancel_qfts:
            # Cancel adjacent QFT gates, which are inverse to each
            # other. This can happen alot because of the heavy use of Fourier arithmetic
            reordered_qc = qft_cancellation(reordered_qc)

        # Transpile logic synthesis
        def logic_synth_transpile_predicate(op):
            return (
                isinstance(op, LogicSynthGate)
                or ("equal" in op.name)
                or ("less_than" in op.name)
                or (op.name in  ["cp"] and op.num_qubits == 2)
                or allocation_level_transpile_predicate(op)
            )

        
        reordered_qc = transpile(
            reordered_qc, transpile_predicate=logic_synth_transpile_predicate
        )

        # We combine adjacent single qubit gates
        if not qs.abstract_params and False:
            reordered_qc = combine_single_qubit_gates(reordered_qc)

        # We now determine the amount of Qubits the circuit will need
        required_qubits = 0
        max_required_qubits = 0

        for instr in reordered_qc.data:
            if instr.op.name == "qb_alloc":
                required_qubits += 1
            elif instr.op.name == "qb_dealloc":
                required_qubits -= 1
            if max_required_qubits < required_qubits:
                max_required_qubits += 1

        # Create a QuantumCircuit. Note that we make sure that the Qubit naming is
        # consistent, since we don't want any non-deterministic elements in the
        # function, as this can hinder bugfixing
        qc = QuantumCircuit()
        for i in range(workspace + max_required_qubits):
            qc.add_qubit(Qubit(str(i)))

        qc.clbits = list(qs.clbits)

        # This dictionary translates between the qubits of the input QuantumSession and
        # the compiled circuit
        translation_dic = {}

        # This list contains the Qubits which are currently not allocated
        free_qb_list = list(qc.qubits)

        allocated_qb_list = []

        # This list contains the clbits used for mcm mcx compilation
        mcm_clbits = []

        depth_dic = {b: 0 for b in qc.qubits + qc.clbits}

        # We now iterate through the data of the preprocessed QuantumCircuit
        data_list = list(reordered_qc.data)

        while data_list:

            instr = data_list.pop(0)

            if instr.op.name == "barrier":
                continue

            # Check if any of the involved qubits need an allocation
            for qb in instr.qubits:
                if qb not in translation_dic:
                    # To pick a good allocation, we determine the depth dic and sort the
                    # available Qubits by their corresponding depth. Note that we add
                    # the identifier in order to prevent non-deterministic behavior

                    alloc_cost = lambda x: (str(depth_dic[x]).zfill(20) + x.identifier)

                    free_qb_list.sort(key=alloc_cost)

                    translation_dic[qb] = free_qb_list.pop(0)

                    allocated_qb_list.append(qb)

            if instr.op.name == "qb_dealloc":
                # For the deallocation, we simply remove the qubits from the translation
                # dict and append it to the free_qb_list
                free_qb_list.append(translation_dic[instr.qubits[0]])
                allocated_qb_list.remove(instr.qubits[0])

                qc.append(
                    instr.op, [translation_dic[qb] for qb in instr.qubits], instr.clbits
                )

                del translation_dic[instr.qubits[0]]

                continue

            if (
                isinstance(instr.op, ControlledOperation)
                and instr.op.base_operation.name == "x"
                and instr.op.method == "auto"
                and len(instr.qubits) > 3
            ):
                # This section deals with the automatic recompilation of mcx gates.
                # The reason why we do this is, that it might not be clear how many free
                # ancillae are available for mcx gates. If we set the method to auto in
                # the mcx function, this section is called and replaces the generated
                # mcx gate with an implementation that is better fit to suit the amount
                # of available ancillae. We first determine the free ancillae

                clean_ancillae = list(free_qb_list)
                clean_ancillae.sort(
                    key=lambda x: depth_dic[x] + qc.qubits.index(x) * 1e-5
                )

                if use_dirty_anc_for_mcx_recomp:
                    dirty_ancillae = list(
                        set(translation_dic.keys()) - set(instr.qubits)
                    )

                    dirty_ancillae.sort(key=lambda x: depth_dic[translation_dic[x]])
                else:
                    dirty_ancillae = []

                QuantumCircuit.fast_append = False
                with fast_append(0):
                    # This function generates the data for the hybrid implementation
                    compiled_mcx_data = gen_hybrid_mcx_data(
                        instr.qubits[:-1],
                        instr.qubits[-1],
                        instr.op.ctrl_state,
                        clean_ancillae,
                        dirty_ancillae,
                        use_mcm=(compile_mcm != None),
                    )

                    data_list = compiled_mcx_data + data_list
                    continue

                    # We now append the data
                    for qb in clean_ancillae + dirty_ancillae:
                        translation_dic[qb] = qb

                    for instr in compiled_mcx_data:
                        qc.append(
                            instr.op, [translation_dic[qb] for qb in instr.qubits]
                        )
                        update_depth_dic(
                            qc.data[-1], depth_dic, depth_indicator=gate_speed
                        )

                    # And free up the qubits
                    for qb in clean_ancillae + dirty_ancillae:
                        del translation_dic[qb]

            elif isinstance(instr.op, GidneyLogicalAND) and compile_mcm:

                if not instr.op.inv:
                    qc.append(
                        instr.op.recompile(compile_mcm),
                        [translation_dic[qb] for qb in instr.qubits],
                    )
                else:

                    compile_qubits = [translation_dic[qb] for qb in instr.qubits]
                    qb_depth = max([depth_dic[qb] for qb in compile_qubits])

                    if len(mcm_clbits) == 0:
                        clbit = qc.add_clbit()
                        depth_dic[clbit] = 0
                        mcm_clbits.append(clbit)

                    min_clbit = np.argmin([depth_dic[cb] for cb in mcm_clbits])

                    if depth_dic[mcm_clbits[min_clbit]] > qb_depth:
                        clbit = qc.add_clbit()
                        depth_dic[clbit] = 0
                        mcm_clbits.append(clbit)
                    else:
                        clbit = mcm_clbits[min_clbit]

                    qc.append(
                        instr.op.recompile(compile_mcm),
                        [translation_dic[qb] for qb in instr.qubits],
                        [clbit],
                    )
                update_depth_dic(qc.data[-1], depth_dic, depth_indicator=gate_speed)

            elif isinstance(instr.op, JonesToffoli) and compile_mcm:
                compile_qubits = [translation_dic[qb] for qb in instr.qubits]
                qb_depth = max([depth_dic[qb] for qb in compile_qubits])

                if len(mcm_clbits) == 0:
                    clbit = qc.add_clbit()
                    depth_dic[clbit] = 0
                    mcm_clbits.append(clbit)
                min_clbit = np.argmin([depth_dic[cb] for cb in mcm_clbits])

                if depth_dic[mcm_clbits[min_clbit]] > qb_depth:
                    clbit = qc.add_clbit()
                    depth_dic[clbit] = 0
                    mcm_clbits.append(clbit)
                else:
                    clbit = mcm_clbits[min_clbit]

                qc.append(
                    instr.op.recompile(),
                    [translation_dic[qb] for qb in instr.qubits],
                    [clbit],
                )
                update_depth_dic(qc.data[-1], depth_dic, depth_indicator=gate_speed)

            # Finally if all of the above cases are not met, we simply append the
            # operation to the translated qubits
            else:
                try:
                    qc.data.append(
                        Instruction(
                            instr.op,
                            [translation_dic[qb] for qb in instr.qubits],
                            instr.clbits,
                        )
                    )

                except KeyError:
                    raise Exception(
                        "Found operation "
                        + instr.op.name
                        + " on unallocated qubit during compilation."
                    )

                update_depth_dic(qc.data[-1], depth_dic, depth_indicator=gate_speed)

        # The following code is mostly about renaming and ordering the resulting circuit
        # in order to make the compiled circuit still comprehensible

        # We rename the allocated qubits to their name from the quantum session
        for i in range(len(qs.qubits)):
            if qs.qubits[i] in translation_dic:
                translation_dic[qs.qubits[i]].identifier = qs.qubits[i].identifier
                translation_dic[qs.qubits[i]].hash_value = qs.qubits[i].hash_value

        # We also want the qubits to be sorted according the order of the
        # QuantumVariables, i.e. the order of creation
        sorted_qubit_list = []
        for qv in qs.qv_list:
            # Due to the measurement reduction feature, not all qubits of quantum
            # variables that are live, are guaranteed to be represented in the
            # translation dict. Therefore, we need the try - except structure here
            try:
                sorted_qubit_list.extend([translation_dic[qb] for qb in qv.reg])
            except KeyError:
                pass

        # Furthermore, not all qubits that have been deallocated in the QuantumSession
        # are guaranteed to be removed from the translation dic, since the
        # measurement_reduction function might have removed their
        # uncomputation/deallocation gates
        temp = list(set(translation_dic.values()) - set(sorted_qubit_list))
        temp.sort(key=lambda x: x.identifier)
        sorted_qubit_list.extend(temp)

        # Finally, we rename the deallocated qubits to "workspace"
        workspace_naming_counter = 0
        td_values = list(translation_dic.values())
        for i in range(len(qc.qubits)):
            if not qc.qubits[i] in td_values:
                qc.qubits[i].identifier = "workspace_" + str(workspace_naming_counter)
                workspace_naming_counter += 1
                sorted_qubit_list.append(qc.qubits[i])

        qc.qubits = sorted_qubit_list

        reduced_qc = parallelize_qc(qc, depth_indicator=gate_speed)
        reduced_qc = cancel_inverses(reduced_qc)
        qc = cancel_inverses(qc)

    if reduced_qc.depth(depth_indicator=gate_speed) > qc.depth(
        depth_indicator=gate_speed
    ):
        return qc
    else:
        return reduced_qc


def gen_hybrid_mcx_data(
    controls, target, ctrl_state, clean_ancillae, dirty_ancillae, use_mcm
):

    # This function generates the data for the hybrid mcx implementation

    from qrisp.core import QuantumVariable
    from qrisp.alg_primitives.mcx_algs import hybrid_mcx

    # Specify QuantumVariables to call mcx function
    control_qv = QuantumVariable(len(controls), name="control")
    target_qv = QuantumVariable(1, name="target")

    # dirty_ancillae = dirty_ancillae + clean_ancillae
    hybrid_mcx(
        control_qv,
        target_qv,
        ctrl_state=ctrl_state,
        num_ancilla=len(clean_ancillae),
        num_dirty_ancilla=len(dirty_ancillae),
        use_mcm=use_mcm,
    )

    # Get the list of used ancillae
    used_ancillae_set = (
        set(control_qv.qs.qubits) - set(control_qv.reg) - set(target_qv.reg)
    )

    # If we used the list() function to transform the set, this introduces a
    # non-deterministic element in the compilation algorithm, which can hamper bugfixing
    used_clean_ancillae = []
    used_dirty_ancillae = []

    for qb in control_qv.qs.qubits:
        if qb in used_ancillae_set:
            if "dirty" in qb.identifier:
                used_dirty_ancillae.append(qb)
            else:
                used_clean_ancillae.append(qb)

    depth_dic = get_depth_dic(control_qv.qs)

    used_clean_ancillae.sort(key=lambda x: -depth_dic[x])
    used_dirty_ancillae.sort(key=lambda x: -depth_dic[x])

    # ancilla_list = list(clean_ancillae)

    # for i in range(len(used_ancillae)):
    #     if "yong" in used_ancillae[i].identifier:
    #         used_ancillae.insert(0, used_ancillae.pop(i))
    #         ancilla_list.insert(0, dirty_ancillae[0])
    #         break

    # Now retarget the instructions such that they use the appropriate qubits
    data = control_qv.qs.data

    i = 0
    # Remove (de)allocation gates
    while i < len(data):
        if (
            data[i].op.name in ["qb_dealloc", "qb_alloc"]
            and not data[i].qubits[0] in used_clean_ancillae
        ):
            data.pop(i)
            continue
        i += 1

    retarget_instructions(data, list(control_qv), controls)
    retarget_instructions(data, list(target_qv), [target])
    # retarget_instructions(data, used_clean_ancillae, clean_ancillae)
    retarget_instructions(data, used_dirty_ancillae, dirty_ancillae)

    return data


# Function to combine any sequences of single qubit gates into a single U3
def combine_single_qubit_gates(qc):
    def apply_combined_gates(qc_new, gate_list, qb):
        if not len(gate_list):
            return

        n = len(gate_list)

        m = np.eye(2)
        while gate_list:
            gate = gate_list.pop(-1)
            m = np.dot(m, gate.get_unitary())

        if np.linalg.norm(m - np.eye(2)) < 1e-10:
            return

        if n == 1:
            qc_new.append(gate, [qb])
            return

        qc_new.unitary(m, [qb])

    qb_dic = {qb: [] for qb in qc.qubits}

    qc_new = qc.clearcopy()

    for instr in qc.data:
        if (
            len(instr.qubits) > 1
            or instr.op.name in ["qb_alloc", "qb_dealloc"]
            or len(instr.clbits) > 0
        ):
            for qb in instr.qubits:
                apply_combined_gates(qc_new, qb_dic[qb], qb)
            qc_new.append(instr)
        else:
            qb_dic[instr.qubits[0]].append(instr.op)

    for qb in qc.qubits:
        apply_combined_gates(qc_new, qb_dic[qb], qb)

    return qc_new


def update_depth_dic(instruction, depth_dic, depth_indicator=None):

    if depth_indicator is None:
        depth_indicator = lambda x: 1

    if instruction.op.definition:
        qc = QuantumCircuit()
        qc.qubits = instruction.qubits
        qc.clbits = instruction.clbits
        qc.append(instruction)
        instr_list = qc.transpile().data
    else:
        instr_list = [instruction]

    # Assign each bit in the circuit a unique integer
    # to index into op_stack.
    # If no bits, return 0

    # Here we are playing a modified version of
    # Tetris where we stack gates, but multi-qubit
    # gates, or measurements have a block for each
    # qubit or cbit that are connected by a virtual
    # line so that they all stacked at the same depth.
    # Conditional gates act on all cbits in the register
    # they are conditioned on.
    # We treat barriers or snapshots different as
    # They are transpiler and simulator directives.
    # The max stack height is the circuit depth.

    for instr in instr_list:
        if instr.op.name in ["qb_alloc", "qb_dealloc", "gphase"]:
            continue
        qargs = instr.qubits
        cargs = instr.clbits

        levels = []
        # If count then add one to stack heights

        for b in qargs + cargs:
            # Add to the stacks of the qubits and
            # cbits used in the gate.
            levels.append(depth_dic[b] + depth_indicator(instr.op))

        max_level = max(levels)

        for b in qargs + cargs:
            depth_dic[b] = max_level


# REWORK required: Instead of detecting a QFT by the gate name, a much more robust
# approach is to create a Operation subtype, that describes QFT gates and do type-
# checking here.


# Function to cancel adjacent QFT, which are inverse to each other
# Due to the heavy use of Fourier arithmetic, this can happen alot
# especially if multiple arithmetic operation on a single target are executed
def qft_cancellation(qc):
    # The idea is to iterate through the instructions of the circuit
    # and save the last instruction acting on each qubit.
    # If a QFT is appended, we check if the last instruction was an inverse QFT
    # or an allocation and denote the index if so
    # Afterwards, we remove the QFTs or replace them with H gates

    last_instr_dic = {qb: None for qb in qc.qubits}
    cancellation_indices = []
    h_replacements = []
    dealloc_replacements = []
    from numpy.linalg import norm

    for i in range(len(qc.data)):

        instr = qc.data[i]
        if "QFT" == instr.op.name[:3] and "adder" not in instr.op.name:
            previous_instruction_type = []

            for qb in instr.qubits:
                previous_instruction = qc.data[last_instr_dic[qb]]

                if previous_instruction.op.name == "qb_alloc":
                    previous_instruction_type.append("alloc")
                    continue

                if previous_instruction.op.num_qubits != instr.op.num_qubits:
                    break

                if not qb in previous_instruction.qubits:
                    break

                if tuple(previous_instruction.qubits) != tuple(instr.qubits):
                    break

                if qc.data.index(previous_instruction) in cancellation_indices:
                    break

                if qc.data.index(previous_instruction) in h_replacements:
                    break

                if "QFT" == previous_instruction.op.name[:3]:
                    if instr.op.num_qubits < 8:
                        unitary_self = instr.op.get_unitary()
                        inv_unitary_other = (
                            previous_instruction.op.get_unitary()
                            .transpose()
                            .conjugate()
                        )

                        if bool(norm(unitary_self - inv_unitary_other) < 10**-4):
                            previous_instruction_type.append("QFT")
                        else:
                            break

                    elif hash(instr.op.definition) == hash(
                        previous_instruction.op.definition.inverse()
                    ):
                        previous_instruction_type.append("QFT")
                    else:
                        break

                else:
                    break

            else:
                if len(set(previous_instruction_type)) == 1:
                    if previous_instruction_type[0] == "QFT":
                        cancellation_indices.append(i)
                        cancellation_indices.append(last_instr_dic[qb])
                    else:
                        h_replacements.append(i)

        if instr.op.name == "qb_dealloc":
            deallocated_qubit = instr.qubits[0]
            previous_instruction = qc.data[last_instr_dic[deallocated_qubit]]

            if "QFT" in previous_instruction.op.name:
                try:
                    previous_instruction.deallocated_qubits[deallocated_qubit] = True
                except AttributeError:
                    previous_instruction.deallocated_qubits = {
                        qb: False for qb in previous_instruction.qubits
                    }
                    previous_instruction.deallocated_qubits[deallocated_qubit] = True
                    dealloc_replacements.append(last_instr_dic[deallocated_qubit])

        for qb in instr.qubits:
            last_instr_dic[qb] = i

    for i in dealloc_replacements:
        instr = qc.data[i]
        for qb in instr.qubits:
            if not instr.deallocated_qubits[qb]:
                break
        else:
            h_replacements.append(i)

    new_qc = qc.clearcopy()

    # print(len(qc.data))
    for i in range(len(qc.data)):
        if i in h_replacements:
            for qb in qc.data[i].qubits:
                new_qc.h(qb)
            # print(qc.data[i].qubits)
            # print("H replacement successfull")
            continue

        if i in cancellation_indices:
            continue

        else:
            new_qc.append(qc.data[i])
    # print(len(new_qc.data))
    # print("====")
    return new_qc


def fuse_instructions(instr_a, instr_b, gphase_array):

    if len(instr_a.qubits) != len(instr_b.qubits):
        return None
    if isinstance(instr_a.op, ClControlledOperation) or isinstance(instr_b.op, ClControlledOperation):
        return None
    
    symmetric_instruction = None
    if instr_a.op.name in ["cz", "swap", "cp", "rzz"]:
        symmetric_instruction = "a"
    if instr_b.op.name in ["cz", "swap", "cp", "rzz"]:
        symmetric_instruction = "b"
    if symmetric_instruction and set(instr_a.qubits) == set(instr_b.qubits):
        pass
    else:
        for i, qb in enumerate(instr_a.qubits):
            if instr_b.qubits[i] != qb:
                return None

    temp = fuse_operations(instr_a.op, instr_b.op, gphase_array)
    if isinstance(temp, Operation):
        if symmetric_instruction is None:
            return Instruction(temp, instr_a.qubits)
        elif symmetric_instruction == "a":
            return Instruction(temp, instr_b.qubits)
        elif symmetric_instruction == "b":
            return Instruction(temp, instr_a.qubits)
    else:
        return temp


def fuse_operations(op_a, op_b, gphase_array):

    from qrisp.alg_primitives import GidneyLogicalAND
    from qrisp.environments import CustomControlOperation
    if "swap" == op_a.name:
        
        if op_b.name == "cx":
            half_swap_qc = QuantumCircuit(2)
            half_swap_qc.cx(0, 1)
            half_swap_qc.cx(1, 0)
            return half_swap_qc.to_gate()
        elif op_b.name == "rzz":
            half_rzz_qc = QuantumCircuit(2)
            half_rzz_qc.cx(0,1)
            half_rzz_qc.cx(1,0)
            half_rzz_qc.rz(op_b.params[0], 1)
            half_rzz_qc.cx(0,1)
            return half_rzz_qc.to_gate()
        elif op_b.name == "cp":
            half_cp_qc = QuantumCircuit(2)
            half_cp_qc.cx(0,1)
            half_cp_qc.cx(1,0)
            half_cp_qc.p(-op_b.params[0]/2, 1)
            half_cp_qc.cx(0,1)
            half_cp_qc.p(op_b.params[0]/2, 0)
            half_cp_qc.p(op_b.params[0]/2, 1)
            return half_cp_qc.to_gate()
    if "swap" == op_b.name:
        if op_a.name == "cx":
            half_swap_qc = QuantumCircuit(2)
            half_swap_qc.cx(1, 0)
            half_swap_qc.cx(0, 1)
            return half_swap_qc.to_gate()
        elif op_a.name == "rzz":
            half_rzz_qc = QuantumCircuit(2)
            half_rzz_qc.cx(0,1)
            half_rzz_qc.cx(1,0)
            half_rzz_qc.rz(op_a.params[0], 1)
            half_rzz_qc.cx(0,1)
            return half_rzz_qc.to_gate()
        elif op_a.name == "cp":
            half_cp_qc = QuantumCircuit(2)
            half_cp_qc.cx(0,1)
            half_cp_qc.cx(1,0)
            half_cp_qc.p(-op_a.params[0]/2, 1)
            half_cp_qc.cx(0,1)
            half_cp_qc.p(op_a.params[0]/2, 0)
            half_cp_qc.p(op_a.params[0]/2, 1)
            return half_cp_qc.to_gate()

    if op_a.definition or op_a.name in ["cx", "cy", "cz"]:

        if isinstance(op_a, ControlledOperation) and isinstance(
            op_b, ControlledOperation
        ):
            if op_a.ctrl_state == op_b.ctrl_state:
                temp = fuse_operations(
                    op_a.base_operation, op_b.base_operation, gphase_array
                )
                if temp == 1:
                    return 1
                elif temp is not None:
                    return ControlledOperation(
                        temp,
                        num_ctrl_qubits=len(op_a.controls),
                        ctrl_state=op_a.ctrl_state,
                    )

        elif isinstance(op_a, GidneyLogicalAND) and isinstance(op_b, GidneyLogicalAND):
            if op_a.ctrl_state == op_b.ctrl_state:
                if op_a.inv != op_b.inv:
                    return 1
            return None

        if isinstance(op_a, CustomControlOperation) and isinstance(
            op_b, CustomControlOperation
        ):
            temp = fuse_operations(
                op_a.definition.data[0].op, op_b.definition.data[0].op, gphase_array
            )
            if temp == 1:
                return 1
            elif temp is not None:
                return CustomControlOperation(temp, op_a.targeting_control)
        return None

    if op_a.params:
        param_sum = sum(op_a.params + op_b.params)
        if op_a.name == "rz":
            if op_b.name == "rz":
                gphase_array[0] += op_a.global_phase + op_b.global_phase
                if param_sum == 0:
                    return 1
                else:
                    return RZGate(param_sum)
            if op_b.name == "p":
                gphase_array[0] += op_a.global_phase
                if param_sum == 0:
                    return 1
                else:
                    return PGate(param_sum)
        elif op_a.name == "p":
            if op_b.name == "rz":
                gphase_array[0] += op_b.global_phase
                if param_sum == 0:
                    return 1
                else:
                    return PGate(param_sum)
            if op_b.name == "p":
                return PGate(param_sum)

        if op_a.name == op_b.name:
            if op_a.name == "rx":
                if param_sum == 0:
                    return 1
                else:
                    return RXGate(param_sum)
            elif op_a.name == "ry":
                if param_sum == 0:
                    return 1
                else:
                    return RYGate(param_sum)
            if op_a.name == "gphase":
                return GPhaseGate(op_a.global_phase + op_b.global_phase)
            else:
                return None

    if op_a.inverse().name == op_b.name and op_a.name[-5:] != "alloc":
        return 1

    return None


def cancel_inverses(qc):
    G = nx.DiGraph()
    qubit_dic = {}
    clbit_dic = {}
    edge_dic = {}

    gphase_array = [0]
    for i in range(qc.num_qubits()):
        qubit_dic[qc.qubits[i]] = -i - 1
        G.add_node(-i - 1)
        
    for i in range(len(qc.clbits)):
        clbit_dic[qc.clbits[i]] = -i-1
        G.add_node(-qc.num_qubits() - i -1)

    data_list = list(qc.data)
    for i in range(len(data_list)):

        G.add_node(i)
        instr = data_list[i]

        predecessors = set()
        for qb in instr.qubits:

            predecessors.add(qubit_dic[qb])

            if not G.has_edge(qubit_dic[qb], i):
                G.add_edge(qubit_dic[qb], i)
                edge_dic[(qubit_dic[qb], i)] = []

            edge_dic[(qubit_dic[qb], i)].append(qb)
            qubit_dic[qb] = i
        
        for cb in instr.clbits:
            G.add_edge(clbit_dic[cb], i)
            clbit_dic[cb] = i

        predecessors = list(predecessors)

        if len(predecessors) == 1:
            pred = predecessors[0]
            if pred < 0:
                continue
            
            fused_gate = fuse_instructions(data_list[pred], data_list[i], gphase_array)
            
            if fused_gate is not None:
                if fused_gate == 1:
                    for edge in G.in_edges(pred):
                        for qb in edge_dic[edge]:
                            qubit_dic[qb] = edge[0]
                        del edge_dic[edge]
                    G.remove_node(pred)
                else:
                    data_list[pred] = fused_gate

                    for edge in G.in_edges(i):
                        for qb in edge_dic[edge]:
                            qubit_dic[qb] = edge[0]
                        del edge_dic[edge]

                G.remove_node(i)

    qc_new = qc.clearcopy()

    topo_sort = list(nx.topological_sort(G))
    with fast_append(3):
        for i in topo_sort:
            if i < 0:
                continue
            else:
                if i == topo_sort[-1]:
                    if data_list[i].op.name == "gphase":
                        gphase_array[0] += data_list[i].op.params[0]
                    if gphase_array[0] != 0:
                        qc_new.gphase(gphase_array[0], data_list[i].qubits[0])
                if data_list[i].op.name != "gphase":
                    qc_new.append(data_list[i].op, data_list[i].qubits, qc.data[i].clbits)
                else:
                    gphase_array[0] += data_list[i].op.params[0]
    return qc_new



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from qrisp.interface import BackendClient


class VirtualBackend(BackendClient):
    """
    This class provides a virtual backend for circuit execution.
    Virtual means that the server is running on the same machine as a separate
    Python thread. This structure allows setting up convenient wrappers for foreign/
    modified circuit dispatching code.

    Circuits can be run using ``virtual_backend_instance.run(qc)``.
    The function that should be used to run a circuit can be specified during
    construction using the ``run_func`` parameter.


    Parameters
    ----------
    run_func : function
        A function that recieves a QuantumCircuit, an integer specifiying the amount of
        shots and a token in the form of a string. It returns the counts as a dictionary
        of bitstrings.
    name : str, optional
        A name for the virtual backend. The default is None.
    port : int, optional
        The port on which to listen. The default is None.

    Examples
    --------


    We set up a VirtualBackend, which prints the received QuantumCircuit and returns the
    results of the QASM simulator. It is required that the run_func specifies a default
    value for the `shots` parameter. ::


        def run_func(qasm_str, shots = 1000, token = ""):

            from qiskit import QuantumCircuit

            qiskit_qc = QuantumCircuit.from_qasm_str(qasm_str)

            print(qiskit_qc)

            from qiskit_aer import AerSimulator
            qiskit_backend = AerSimulator()

            #Run Circuit on the Qiskit backend
            return qiskit_backend.run(qiskit_qc, shots = shots).result().get_counts()


    >>> from qrisp.interface import VirtualBackend
    >>> example_backend = VirtualBackend(run_func)
    >>> from qrisp import QuantumFloat
    >>> qf = QuantumFloat(3)
    >>> qf[:] = 4
    >>> qf.get_measurement(backend = example_backend)
                 ┌─┐
    qf.0:   ─────┤M├──────
                 └╥┘┌─┐
    qf.1:   ──────╫─┤M├───
            ┌───┐ ║ └╥┘┌─┐
    qf.2:   ┤ X ├─╫──╫─┤M├
            └───┘ ║  ║ └╥┘
    cb_0: 1/══════╩══╬══╬═
                  0  ║  ║
    cb_1: 1/═════════╩══╬═
                     0  ║
    cb_2: 1/════════════╩═
    {4: 1.0}

    """

    def __init__(self, run_func, port=None):

        from qrisp.interface import BackendServer

        self.port = port
        if port is None:
            self.run_func = run_func
        else:
            # Create BackendServer
            self.backend_server = BackendServer(run_func, "localhost", port=port)
            # Run the server (runs in the background)
            self.backend_server.start()
            # Connect client

            super().__init__(api_endpoint="localhost", port=port)

    def run(self, qc, shots=None, token=""):
        """
        Executes the function run_func specified at object creation.

        Parameters
        ----------
        qc : QuantumCircuit
            The QuantumCircuit to run.
        shots : int, optional
            The amount of shots to perform.

        Returns
        -------
        res : dict
            A dictionary containing the measurement results.

        """
        if self.port is None:
            return self.run_func(qc.qasm(), shots, token)
        else:
            return super().run(qc, shots)



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import jax
import jax.numpy as jnp
from jax.lax import fori_loop
from jax.scipy.optimize import OptimizeResults


# https://www.jhuapl.edu/SPSA/PDF-SPSA/Spall_An_Overview.PDF
# Conditions: alpha <= 1; 1/6 <= gamma <= 1/2; 2*(alpha-gamma) > 1
def spsa(fun, x0, args, maxiter=50, a=2.0, c=0.1, alpha=0.702, gamma=0.201, seed=3):
    r"""

    Minimize a scalar function of one or more variables using the `Simultaneous Perturbation Stochastic Approximation algorithm <https://en.wikipedia.org/wiki/Simultaneous_perturbation_stochastic_approximation>`_.

    This algorithm aims at finding the optimal control $x^*$ minimizing a given loss fuction $f$:

    .. math::

        x^* = \text{argmin}_{x} f(x)

    This is done by an iterative process starting from an initial guess $x_0$:

    .. math::

        x_{k+1} = x_k - a_kg_k(x_k)

    where $a_k=\dfrac{a}{n^{\alpha}}$ for scaling parameters $a, \alpha>0$.

    For each step $x_k$ the gradient is approximated by

    .. math::

        (g_k(x_k))_i = \frac{f(x_k+c_k\Delta_k)-f(x_k-c_k\Delta_k)}{2c_k(\Delta_k)_i}

    where $c_k=\dfrac{c}{n^{\gamma}}$ for scaling parameters $c, \gamma>0$, and $\Delta_k$ is a random perturbation vector.

    Parameters
    ----------
        maxiter : int
            Maximum number of iterations to perform. Each iteration requires 2 function evaluations.
        a : float
            Scaling parameter for update rule.
        alpha : float
            Scaling exponent for update rule.
        c : float
            Scaling parameter for gradient estimation.
        gamma : float
            Scaling exponent for gradient estimation.

    Returns
    -------
    results
        An `OptimizeResults <https://docs.jax.dev/en/latest/_autosummary/jax.scipy.optimize.OptimizeResults.html#jax.scipy.optimize.OptimizeResults>`_ object.

    """

    rng = jax.random.PRNGKey(seed)

    def body_fun(k, state):

        x, rng = state

        # Generate random perturbation delta with components +/-1
        rng, rng_input = jax.random.split(rng)
        delta = jax.random.choice(rng, jnp.array([1, -1]), shape=(*x.shape,))

        ak = a / (k + 1) ** alpha
        ck = c / (k + 1) ** gamma

        # Evaluate loss function at perturbed points
        x_plus = x + ck * delta
        x_minus = x - ck * delta

        loss_plus = fun(x_plus, *args)
        loss_minus = fun(x_minus, *args)

        # Approximate gradient
        gk = (loss_plus - loss_minus) / (2.0 * ck * delta)

        # Update parameters
        x = x - ak * gk

        return x, rng

    from qrisp.jasp import make_tracer

    x, rng = fori_loop(0, make_tracer(maxiter), body_fun, (x0, rng))
    fx = fun(x, *args)

    return OptimizeResults(x, True, 0, fx, None, None, 2 * maxiter + 1, 0, maxiter)



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import numpy as np
from sympy import continued_fraction_convergents, continued_fraction_iterator, Rational

from qrisp.interface import QiskitBackend
from qrisp.alg_primitives.arithmetic.modular_arithmetic import find_optimal_m, modinv
from qrisp.alg_primitives import QFT
from qrisp import QuantumModulus, QuantumFloat, h, control

depths = []
cnot_count = []
qubits = []


def find_optimal_a(N):
    n = int(np.ceil(np.log2(N)))
    proposals = []

    # Search through the first O(1) possibilities to find a good a
    for a in range(2, min(100, N - 1)):
        # We only append non-trivial proposals
        if np.gcd(a, N) == 1:
            proposals.append(a)

    cost_dic = {}
    for a in proposals:
        m_values = []
        for k in range(2 * n + 1):
            inpl_multiplier = (a ** (2**k)) % N

            if inpl_multiplier == 1:
                continue

            # find_optimal_m is a function that determines the lowest possible
            # Montgomery shift for a given number. The higher the montgomery shift,
            # the more qubits and the more effort is needed.
            m_values.append(find_optimal_m(inpl_multiplier, N))
            m_values.append(find_optimal_m(modinv((-inpl_multiplier) % N, N), N))

        cost_dic[a] = sum(m_values) + max(m_values) * 1e-5

    proposals.sort(key=lambda a: cost_dic[a])

    optimal_a = proposals[0]

    m_values = []

    for k in range(2 * n + 1):
        inpl_multiplier = ((optimal_a) ** (2**k)) % N

        if inpl_multiplier == 1:
            continue

        m_values.append(find_optimal_m(inpl_multiplier, N))

    return proposals


def find_order(a, N, inpl_adder=None, mes_kwargs={}):
    qg = QuantumModulus(N, inpl_adder)
    qg[:] = 1
    qpe_res = QuantumFloat(2 * qg.size + 1, exponent=-(2 * qg.size + 1))
    h(qpe_res)
    for i in range(len(qpe_res)):
        with control(qpe_res[i]):
            qg *= a
            a = (a * a) % N
    QFT(qpe_res, inv=True, inpl_adder=inpl_adder)

    mes_res = qpe_res.get_measurement(**mes_kwargs)

    return extract_order(mes_res, a, N)


def extract_order(mes_res, a, N):

    collected_r_values = []

    approximations = list(mes_res.keys())

    try:
        approximations.remove(0)
    except ValueError:
        pass

    while True:

        r_values = get_r_values(approximations.pop(0))

        for r in r_values:
            if (a**r) % N == 1:
                return r

        collected_r_values.append(r_values)
        from itertools import product

        for comb in product(*collected_r_values):
            r = np.lcm.reduce(comb)
            if (a**r) % N == 1:
                return r


def get_r_values(approx):
    rationals = continued_fraction_convergents(
        continued_fraction_iterator(Rational(approx))
    )
    return [rat.q for rat in rationals if 1 < rat.q]


def shors_alg(N, inpl_adder=None, mes_kwargs={}):
    """
    Performs `Shor's factorization algorithm <https://arxiv.org/abs/quant-ph/9508027>`_ on a given integer N.
    The adder used for factorization can be customized. To learn more about this feature, please read :ref:`QuantumModulus`

    Parameters
    ----------
    N : integer
        The integer to be factored.
    inpl_adder : callable, optional
        A function that performs in-place addition. The default is None.
    mes_kwargs : dict, optional
        A dictionary of keyword arguments for :meth:`get_measurement <qrisp.QuantumVariable.get_measurement>`. This especially allows you to specify an execution backend. The default is {}.

    Returns
    -------
    res : integer
        A factor of N.

    Examples
    --------

    We factor 65:

    >>> from qrisp.shor import shors_alg
    >>> shors_alg(65)
    5

    """
    if not N % 2:
        return 2

    a_proposals = find_optimal_a(N)

    for a in a_proposals:

        K = np.gcd(a, N)

        if K != 1:
            res = K
            break

        r = find_order(a, N, inpl_adder, mes_kwargs)

        if r % 2:
            continue

        g = int(np.gcd(a ** (r // 2) + 1, N))

        if g not in [N, 1]:
            res = g
            break
    return res



"""
********************************************************************************
* Copyright (c) 2024 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""


def test_QMCI():
    from qrisp import QuantumFloat
    from qrisp.algorithms.qmci import QMCI
    import numpy as np

    def f(qf):
        return qf*qf

    qf_x = QuantumFloat(3,-3)
    qf_y = QuantumFloat(6,-6)
    a = QMCI([qf_x,qf_y], f)

    assert np.abs(a-0.2734375)<0.01


"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from qrisp import h, control, rz, measure, QuantumFloat
from qrisp.jasp import jrange, q_fori_loop
import numpy as np


def IQPE(args, U, precision, iter_spec=False, ctrl_method=None, kwargs={}):
    r"""
    Evaluates the `iterative quantum phase estimation algorithm
    <https://arxiv.org/pdf/quant-ph/0610214>`_.

    The unitary to estimate is expected to be given as Python function, which is called
    on ``args``.

    Parameters
    ----------
    args : list
        A list of arguments (could be QuantumVariables) which represent the state,
        the quantum phase estimation is performed on.
    U : function
        A Python function, which will receive the list ``args`` as arguments in the
        course of this algorithm.
    precision : int
        The precision of the estimation.
    iter_spec : bool, optional
        If set to ``True``, ``U`` will be called with the additional keyword
        ``iter = i`` where ``i`` is the amount of iterations to perform (instead of
        simply calling ``U`` for ``i`` times). The default is False.
    ctrl_method : string, optional
        Allows to specify which method should be used to generate the
        controlled U circuit. For more information check
        :meth:`.control <qrisp.Operation.control>`. The default is None.
    kwargs : dict, optional
        A dictionary of keyword arguments to pass to ``U``. The default is {}.

    Returns
    -------
    theta : float
        The estimated phase as a fraction of $2 \pi$.

    Examples
    --------

    We define a function that applies two rotations onto its input and estimate the
    applied phase. ::

        from qrisp import IQPE, h, run, x, rx, QuantumFloat
        from qrisp.jasp import make_jaspr
        import numpy as np

        def f():
            def U(qv):
                x = 1/2**3
                y = 1/2**2

                rx(x*2*np.pi, qv[0])
                rx(y*2*np.pi, qv[1])

            qv = QuantumFloat(2)

            x(qv)
            h(qv)

            return IQPE(qv, U, precision = 4)
        jaspr = make_jaspr(f)()

    >>> jaspr()
    Array(0.375, dtype=float64)

    """

    def iqpe_iteration(i, val):
        theta, args = val
        iqpe_aux = QuantumFloat(1)
        h(iqpe_aux)
        if iter_spec:
            with control(iqpe_aux[0], ctrl_method=ctrl_method):
                U(args, iter=2 ** (precision - i), **kwargs)
        else:
            with control(iqpe_aux[0], ctrl_method=ctrl_method):
                for _ in jrange(2 ** (precision - i)):
                    U(args, **kwargs)
        rz(-np.pi * theta, iqpe_aux)
        h(iqpe_aux)
        r = measure(iqpe_aux)
        iqpe_aux.delete()
        return (theta + r) / 2, args

    return q_fori_loop(0, precision, iqpe_iteration, (0, args))[0]



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import jax.numpy as jnp

from qrisp.core import QuantumVariable, cx
from qrisp.qtypes import QuantumBool
from qrisp.environments import invert, adaptive_condition, conjugate, control


def uint_qq_less_than(a, b, inv_adder):
    comparison_anc = QuantumBool()
    comparison_res = QuantumBool()

    temp_var = QuantumVariable(jnp.maximum(0, b.size - a.size))
    a = a.ensure_reg() + temp_var.ensure_reg()

    with conjugate(inv_adder, allocation_management=False)(b, a + comparison_anc.reg):
        cx(comparison_anc, comparison_res)

    comparison_anc.delete()

    temp_var.delete()

    return comparison_res


def uint_cq_less_than(a, b, inv_adder):

    comparison_res = QuantumBool()

    with control(a < 0):
        comparison_res.flip()

    with control((a >= 0) & (a < 2**b.size - 1)):
        comparison_anc = QuantumBool()

        with conjugate(inv_adder, allocation_management=False)(
            a + 1, b.reg + comparison_anc.reg
        ):
            cx(comparison_anc, comparison_res)

        comparison_res.flip()
        comparison_anc.delete()

    return comparison_res


def uint_qc_less_than(a, b, inv_adder):

    comparison_res = QuantumBool()

    with control(b > 2**a.size):
        comparison_res.flip()
    with control((b >= 0) & (b <= 2**a.size)):
        comparison_anc = QuantumBool()

        with conjugate(inv_adder, allocation_management=False)(
            b, a.reg + comparison_anc.reg
        ):
            cx(comparison_anc, comparison_res)

        comparison_anc.delete()

    return comparison_res


def uint_less_than(a, b, inpl_adder):

    def inv_adder(a, b):
        with invert():
            inpl_adder(a, b)

    if isinstance(a, QuantumVariable) and isinstance(b, QuantumVariable):
        return uint_qq_less_than(a, b, inv_adder)
    elif isinstance(a, QuantumVariable):
        return uint_qc_less_than(a, b, inv_adder)
    elif isinstance(b, QuantumVariable):
        return uint_cq_less_than(a, b, inv_adder)
    else:
        return a < b


def uint_lt(a, b, inpl_adder):
    return uint_less_than(a, b, inpl_adder)


def uint_gt(a, b, inpl_adder):
    return uint_less_than(b, a, inpl_adder)


def uint_le(a, b, inpl_adder):
    return uint_less_than(b, a, inpl_adder).flip()


def uint_ge(a, b, inpl_adder):
    return uint_less_than(a, b, inpl_adder).flip()



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from functools import lru_cache
from random import shuffle

import numpy as np

import jax.numpy as jnp
import jax

from qrisp.jasp.interpreter_tools.abstract_interpreter import (
    eval_jaxpr,
    extract_invalues,
    insert_outvalues,
    exec_eqn,
)
from qrisp.jasp.interpreter_tools.interpreters.control_flow_interpretation import (
    evaluate_while_loop,
)

# The following function implements the behavior of the jaspify simulator for terminal sampling
# To understand the function consider the result of tracing a simple sampling task

# from qrisp.jasp import *
# from qrisp import *

# def inner_f():
#     qf = QuantumFloat(4)
#     h(qf[0])
#     return qf

# @make_jaxpr
# def main(i):

#     return expectation_value(inner_f, i)()

# print(main(100))

# While the resulting Jaxpr accurately describes a sampling task, the sampling
# is not as scalable as it could be, given the powerfull sampling features of the
# Qrisp simulator.

# When called via @jaspify and terminal_sampling = True, we need some custom logic
# to implement the terminal_sampling behavior.

# The traced Jaxpr contains the "sampling_body_func", which
# is (visually) partioned into three sections:

# 1. The evaluation of the user function, which executes the code of "inner_func"
# 2. The sampling helper 1 function. This function contains the measurement primitives.
# 3. The sampling helper 2 function. This function contains the decoding of the measurement results.

# These functions are evaluated with custom code. Of particular importance is step 2,
# which leverages the fast sampling features of the Qrisp simulator (instead of looping)
# over measurements.

# { lambda a:f64[1]; b:i64[]. let
#     c:f64[] = pjit[
#       name=expectation_value_eval_function
#       jaxpr={ lambda ; d:f64[1] e:i64[]. let
#           _:i64[] _:i64[] f:f64[1] = while[
#             body_jaxpr={ lambda ; g:i64[] h:i64[] i:f64[1]. let
#                 j:i64[] = add g 1
#                 k:QuantumCircuit = jasp.quantum_kernel
#                 _:QuantumCircuit l:f64[1] = pjit[
#                   name=sampling_body_func
#                   jaxpr={ lambda ; m:QuantumCircuit n:i64[] o:f64[1]. let

# -------------------------------------------------------------------------------------

#                       p:QuantumCircuit q:QubitArray r:i64[] _:bool[] = pjit[
#                         name=user_func
#                         jaxpr={ lambda ; s:QuantumCircuit. let
#                             t:QuantumCircuit u:QubitArray = jasp.create_qubits s
#                               4
#                             v:Qubit = jasp.get_qubit u 0
#                             w:QuantumCircuit = jasp.h t v
#                           in (w, u, 0, False) }
#                       ] m

# -------------------------------------------------------------------------------------

#                       x:QuantumCircuit y:i64[] = pjit[
#                         name=sampling_helper_1
#                         jaxpr={ lambda ; z:QuantumCircuit ba:QubitArray. let
#                             bb:QuantumCircuit bc:i64[] = jasp.measure z ba
#                           in (bb, bc) }
#                       ] p q

# -------------------------------------------------------------------------------------

#                       bd:f64[] = pjit[
#                         name=sampling_helper_2
#                         jaxpr={ lambda ; be:i64[] bf:i64[]. let
#                             _:f64[] = convert_element_type[
#                               new_dtype=float64
#                               weak_type=False
#                             ] be
#                             bg:f64[] = pow 2.0 be
#                             bh:f64[] = convert_element_type[
#                               new_dtype=float64
#                               weak_type=False
#                             ] bf
#                             bi:f64[] = mul bh bg
#                           in (bi,) }
#                       ] r y

# -------------------------------------------------------------------------------------

#                       bj:f64[1] = broadcast_in_dim[
#                         broadcast_dimensions=()
#                         shape=(1,)
#                       ] bd
#                       bk:f64[1] = add o bj
#                       bl:QuantumCircuit = jasp.reset x q
#                       _:QuantumCircuit = jasp.delete_qubits bl q
#                     in (x, bk) }
#                 ] k g i
#               in (j, h, l) }
#             body_nconsts=0
#             cond_jaxpr={ lambda ; bm:i64[] bn:i64[] bo:f64[1]. let
#                 bp:bool[] = lt bm bn
#               in (bp,) }
#             cond_nconsts=0
#           ] 0 e d
#           bq:f64[1] = slice[limit_indices=(1,) start_indices=(0,) strides=None] f
#           br:f64[] = squeeze[dimensions=(0,)] bq
#           bs:f64[] = convert_element_type[new_dtype=float64 weak_type=False] e
#           bt:f64[] = div br bs
#         in (bt,) }
#     ] a b
#   in (c,) }


def terminal_sampling_evaluator(sampling_res_type):

    def sampling_eqn_evaluator(eqn, context_dic, eqn_evaluator=exec_eqn):

        invalues = extract_invalues(eqn, context_dic)

        # Extract the shot number
        if sampling_res_type == "ev":
            shots = invalues[-1]
        elif sampling_res_type == "array":
            shots = invalues[-1]
        elif sampling_res_type == "dict":
            shots = invalues[-1]
            if shots == 0:
                invalues[-1] = 1

        # We will use this dictionary to store the sampling results
        meas_res_dic = {}

        # This list will contain information about how many qubits are contained
        # in each QuantumVariable of the returned function
        return_signature = []

        decoded_meas_res = []

        def sampling_body_eqn_evaluator(eqn, context_dic):

            # We only do a single iteration (the actual iteration is performed within)
            # the simulator.
            if eqn.primitive.name == "while":
                return evaluate_while_loop(
                    eqn,
                    context_dic,
                    eqn_evaluator=sampling_body_eqn_evaluator,
                    break_after_first_iter=True,
                )

            if eqn.primitive.name == "pjit":
                invalues = extract_invalues(eqn, context_dic)

                function_name = eqn.params["name"]

                # sampling_body_func is called with the ev_eqn_evaluator
                if function_name == "sampling_body_func":
                    outvalues = eval_jaxpr(
                        eqn.params["jaxpr"], eqn_evaluator=sampling_body_eqn_evaluator
                    )(*invalues)
                    insert_outvalues(eqn, context_dic, outvalues)
                    return

                if function_name == "user_func":
                    pass

                # This case describes the logic to use the simulator sampling features
                if function_name == "sampling_helper_1":

                    # Collect the qubits to be measured into a single list
                    qubits = []
                    for i in range(len(invalues) - 1):
                        qubits.extend(invalues[i])
                        return_signature.append(len(invalues[i]))

                    # Clear the buffer and copy the state to evaluate the measurement
                    invalues[-1].apply_buffer()
                    quantum_state = invalues[-1].copy()

                    # Evaluate the measurement
                    # This returns a dictionary of the form {label int : count int}
                    # if shots is an int. If shots is None, count int instead is a
                    # float representing the probability.

                    meas_res_dic.update(quantum_state.multi_measure(qubits, shots))

                    if shots is None:
                        # Round to prevent floating point errors of the simulation
                        norm = 0
                        for k, v in meas_res_dic.items():
                            meas_res_dic[k] = np.round(v, decimals=7)
                            norm += meas_res_dic[k]

                        for k, v in meas_res_dic.items():
                            meas_res_dic[k] = v / norm

                    outvalues = [0] * len(eqn.outvars)
                    insert_outvalues(eqn, context_dic, outvalues)

                # Each int in the values of meas_res_dic represents the values
                # of all QuantumVariables. We therefore need to "split" the ints
                # into the appropriate parts and decode them.
                # Splitting means turning the int "1001001" into "100" and "1001".
                if function_name == "sampling_helper_2":

                    if sampling_res_type == "ev":
                        sampling_res = jnp.zeros(len(eqn.outvars))
                    elif sampling_res_type == "array":
                        sampling_res = []
                        sampling_res_dict = {}
                    elif sampling_res_type == "dict":
                        sampling_res = {}

                    # Compile the decoder
                    decoder = decoder_compiler(eqn.params["jaxpr"], eqn_evaluator)

                    # Iterate through the sampled values
                    for k, v in meas_res_dic.items():

                        # We now evaluate the function that was previously traced
                        # to perform the decoding. The first few arguments of this
                        # function are the integers to be decoded.

                        # We therefore use the traced Jaspr to perform the decoding
                        # by modifying the input values.
                        new_invalues = list(invalues)

                        j = 0
                        for i in range(len(return_signature)):
                            # Split the integers into intervals ranging from
                            # j to j + return_signature[i]
                            new_invalues[len(invalues) - len(return_signature) + i] = (
                                k & ((2 ** (return_signature[i]) - 1) << j)
                            ) >> j
                            j += return_signature[i]

                        # Evaluate the decoder
                        # outvalues = eval_jaxpr(eqn.params["jaxpr"], eqn_evaluator = eqn_evaluator)(*new_invalues)

                        outvalues = decoder(*new_invalues)

                        # We now build the key for the result dic
                        # For that we turn the jax types into the corresponding
                        # Python types.
                        if not isinstance(outvalues, tuple):
                            if sampling_res_type == "ev":
                                sampling_res += outvalues * v
                            elif sampling_res_type == "array":
                                # sampling_res.extend(v*[outvalues[0]])
                                key = outvalues[0]
                                if key.size == 1:
                                    sampling_res_dict[key.item()] = v
                                else:
                                    sampling_res_dict[tuple(np.array(key))] = v
                                    # sampling_res.extend(v*[key])
                            elif sampling_res_type == "dict":
                                key = outvalues
                                if not type(v) in [int, float]:
                                    if v.dtype in [np.float64, np.float32]:
                                        v = float(v.item())
                                    elif v.dtype in [np.int32, np.int64]:
                                        v = int(v.item())
                                    else:
                                        raise
                                sampling_res[key.item()] = v

                        # If the user given function returned more than one
                        # value, the key is a tuple to be build up
                        else:
                            if sampling_res_type == "ev":
                                sampling_res += jnp.array(outvalues) * v
                            elif sampling_res_type == "array":
                                sampling_res.extend(v * [outvalues])
                            elif sampling_res_type == "dict":
                                if not type(v) in [int, float]:
                                    if v.dtype in [np.float64, np.float32]:
                                        v = float(v.item())
                                    elif v.dtype in [np.int32, np.int64]:
                                        v = int(v.item())
                                    else:
                                        raise
                                sampling_res[tuple(x.item() for x in outvalues)] = v

                    if sampling_res_type == "array":
                        keys = np.array(list(sampling_res_dict.keys()))
                        counts = np.array(list(sampling_res_dict.values()))
                        sampling_res = np.repeat(keys, counts, axis=0)
                        np.random.shuffle(sampling_res)

                    elif sampling_res_type == "ev":
                        sampling_res = sampling_res / shots
                        if sampling_res.shape[0] == 1:
                            sampling_res = sampling_res[0]
                    elif sampling_res_type == "dict":
                        # Sort the counts such the most probable values come first
                        sampling_res = dict(
                            sorted(sampling_res.items(), key=lambda item: item[0])
                        )
                        sampling_res = dict(
                            sorted(sampling_res.items(), key=lambda item: -item[1])
                        )

                    decoded_meas_res.append(sampling_res)

            return eqn_evaluator(eqn, context_dic)

        # Execute the above defined interpreter
        sampling_body_jaxpr = eqn.params["jaxpr"].jaxpr

        outvalues = eval_jaxpr(
            sampling_body_jaxpr, eqn_evaluator=sampling_body_eqn_evaluator
        )(*invalues)

        if not isinstance(outvalues, (list, tuple)):
            outvalues = [outvalues]

        insert_outvalues(eqn, context_dic, decoded_meas_res)

    return sampling_eqn_evaluator


@lru_cache(maxsize=int(1e5))
def decoder_compiler(jaxpr, eqn_evaluator):
    return jax.jit(eval_jaxpr(jaxpr, eqn_evaluator=eqn_evaluator))



"""
********************************************************************************
* Copyright (c) 2024 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from qrisp import QuantumVariable   
from qrisp.operators import X, Y, Z
from qrisp.vqe.problems.heisenberg import create_heisenberg_init_function
from qrisp.qite import QITE
import networkx as nx
import numpy as np
import sympy as sp

def test_qite_heisenberg():

    # Create a graph
    N = 4
    G = nx.Graph()
    G.add_edges_from([(k,k+1) for k in range(N-1)]) 

    def create_heisenberg_hamiltonian(G):
        H = sum(X(i)*X(j)+Y(i)*Y(j)+Z(i)*Z(j) for (i,j) in G.edges())
        return H

    H = create_heisenberg_hamiltonian(G)
    E_tar = H.ground_state_energy()

    M = nx.maximal_matching(G)
    U_0 = create_heisenberg_init_function(M)

    qv = QuantumVariable(N)
    U_0(qv)
    E_0 = H.get_measurement(qv)

    def exp_H(qv, t):
        H.trotterization(method='commuting')(qv,t,5)
    
    steps = 4
    s_values = np.linspace(.01,.3,10)

    theta = sp.Symbol('theta')
    optimal_s = [theta]
    optimal_energies = [E_0]

    for k in range(1,steps+1):

        # Perform k steps of QITE
        qv = QuantumVariable(N)
        QITE(qv, U_0, exp_H, optimal_s, k)
        qc = qv.qs.compile()

        # Find optimal evolution time 
        energies = [H.get_measurement(qv,subs_dic={theta:s_},precompiled_qc=qc,diagonalisation_method='commuting') for s_ in s_values]
        index = np.argmin(energies)
        s_min = s_values[index]

        optimal_s.insert(-1,s_min)
        optimal_energies.append(energies[index])

    assert np.abs(E_tar-optimal_energies[-1]) < 1e-1


"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import numpy as np

from qrisp.qtypes.quantum_float import QuantumFloat
from qrisp.qtypes.quantum_bool import QuantumBool
from qrisp.alg_primitives.arithmetic import multi_controlled_U_g, hybrid_mult, U_g
from qrisp.alg_primitives import QFT
from qrisp.core.gate_application_functions import h, cx, swap
from qrisp.environments import conjugate, control, invert, custom_control
from qrisp.circuit import Operation
from qrisp.alg_primitives.arithmetic.modular_arithmetic.mod_tools import (
    modinv,
    montgomery_decoder,
    montgomery_encoder,
)


def qft_basis_adder(addend, target):

    if isinstance(addend, int):
        U_g(addend, target)
    elif isinstance(addend, QuantumFloat):
        if addend.signed:
            raise Exception("Signed addition not supported")
        for i in range(*addend.mshape):
            multi_controlled_U_g(target, [addend.significant(i)], 2**i)


# Performs the modular inplace addition b += a
# where a and b don't need to have the same montgomery shift
def montgomery_addition(a, b):

    for i in range(len(a)):
        with control(a[i]):
            b += pow(2, i - a.m, a.modulus)


def beauregard_adder(a, b, modulus):

    if modulus > 2**a.size:
        raise Exception(
            "Tried to perform modular addition on QuantumFloat with too few qubits"
        )
    if modulus == 2**a.size:
        with conjugate(QFT)(a, exec_swap=False):
            qft_basis_adder(b, a)
        return

    reduction_not_necessary = QuantumBool()
    sign = QuantumBool()

    if isinstance(b, int):
        b = b % modulus

    a = list(a) + [sign[0]]

    with conjugate(QFT)(a, exec_swap=False):

        qft_basis_adder(b, a)

        with invert():
            qft_basis_adder(modulus, a)

        with conjugate(QFT)(a, exec_swap=False, inv=True):
            cx(sign, reduction_not_necessary)

        with control(reduction_not_necessary):
            qft_basis_adder(modulus, a)

        with invert():
            qft_basis_adder(b, a)

        with conjugate(QFT)(a, exec_swap=False, inv=True):
            cx(sign, reduction_not_necessary)
            reduction_not_necessary.flip()

        qft_basis_adder(b, a)

    sign.delete()
    reduction_not_necessary.delete()


@custom_control
def mod_adder(a, b, inpl_adder, modulus, ctrl=None):

    reduction_not_necessary = QuantumBool()
    sign = QuantumBool()

    if isinstance(a, int):
        a = a % modulus

    b = list(b) + [sign[0]]

    if ctrl is None:
        inpl_adder(a, b)
    else:
        with control(ctrl):
            inpl_adder(a, b)

    with invert():
        inpl_adder(modulus, b)

    cx(sign, reduction_not_necessary)

    with control(reduction_not_necessary):
        inpl_adder(modulus, b)

    with invert():
        if ctrl is None:
            inpl_adder(a, b)
        else:
            with control(ctrl):
                inpl_adder(a, b)

    cx(sign, reduction_not_necessary)
    reduction_not_necessary.flip()

    if ctrl is None:
        inpl_adder(a, b)
    else:
        with control(ctrl):
            inpl_adder(a, b)

    sign.delete()
    reduction_not_necessary.delete()



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import numpy as np

from qrisp import QuantumCircuit, CZGate, XGate, Instruction


def ctrl_state_wrap(qc, ctrl_state):
    res = qc.copy()
    for i in range(len(ctrl_state)):
        if ctrl_state[i] == "0":
            res.data.insert(0, Instruction(XGate(), [res.qubits[i]]))
            res.x(res.qubits[i])
    return res


# Toffoli Implementation according to https://arxiv.org/pdf/1206.0758.pdf
toffoli_qc = QuantumCircuit(3)
toffoli_qc.h(2)
toffoli_qc.t_dg([0, 1])
toffoli_qc.cx(2, 0)
toffoli_qc.cx(1, 2)
toffoli_qc.t(0)
toffoli_qc.cx(1, 0)
toffoli_qc.t(2)
toffoli_qc.cx(1, 2)
toffoli_qc.t_dg(0)
toffoli_qc.cx(2, 0)
toffoli_qc.t(0)
toffoli_qc.t_dg(2)
toffoli_qc.cx(1, 0)
toffoli_qc.h(2)
toffoli_qc.qubits[1], toffoli_qc.qubits[0] = toffoli_qc.qubits[0], toffoli_qc.qubits[1]


# Margolus gate as described here https://arxiv.org/abs/quant-ph/0312225 and
# here https://www.iccs-meeting.org/archive/iccs2022/papers/133530169.pdf
margolus_qc = QuantumCircuit(3)

margolus_qc.sx(2)
margolus_qc.t(2)
margolus_qc.cx(1, 2)
margolus_qc.t(2)
margolus_qc.cx(0, 2)

reduced_margolus_qc = margolus_qc.copy()
reduced_margolus_qc.sx_dg(2)

margolus_qc.t_dg(2)
margolus_qc.cx(1, 2)
margolus_qc.t_dg(2)
margolus_qc.sx_dg(2)

# Ancilla supported multi controlled X gates from https://arxiv.org/pdf/1508.03273.pdf
# and https://www.iccs-meeting.org/archive/iccs2022/papers/133530169.pdf
maslov_qc = QuantumCircuit(4)

maslov_qc.h(3)
maslov_qc.t(3)
maslov_qc.cx(2, 3)
maslov_qc.t_dg(3)
maslov_qc.h(3)
maslov_qc.cx(0, 3)
maslov_qc.t(3)
maslov_qc.cx(1, 3)
maslov_qc.t_dg(3)
maslov_qc.cx(0, 3)

reduced_maslov_qc = maslov_qc.copy()

maslov_qc.t(3)
maslov_qc.cx(1, 3)
maslov_qc.t_dg(3)
maslov_qc.h(3)
maslov_qc.t(3)
maslov_qc.cx(2, 3)
maslov_qc.t_dg(3)
maslov_qc.h(3)


# Quasi-Toffoli implementation from https://arxiv.org/abs/1709.06648
gidney_qc = QuantumCircuit(3)

gidney_qc.h(2)
gidney_qc.t(2)
gidney_qc.cx([0, 1], 2)
gidney_qc.cx(2, [0, 1])
gidney_qc.t_dg([0, 1])
gidney_qc.t(2)
gidney_qc.cx(2, [0, 1])
gidney_qc.h([2])
gidney_qc.s([2])

gidney_qc_inv = QuantumCircuit(3, 1)

gidney_qc_inv.h(2)
gidney_qc_inv.measure(2, 0)
gidney_qc_inv.append(CZGate().c_if(1), [0, 1], 0)
gidney_qc_inv.reset(2)

# Implementation of the T-depth 2 Toffoli found in https://arxiv.org/pdf/1206.0758.pdf
amy_toffoli_qc = QuantumCircuit(4)
amy_toffoli_qc.h(2)
amy_toffoli_qc.t([0, 1, 2])
amy_toffoli_qc.cx(1, 0)
amy_toffoli_qc.cx(2, 1)
amy_toffoli_qc.cx(0, 2)
amy_toffoli_qc.cx(0, 3)
amy_toffoli_qc.cx(1, 3)
amy_toffoli_qc.t_dg([0, 1, 3])
amy_toffoli_qc.t([2])
amy_toffoli_qc.cx(1, 3)
amy_toffoli_qc.cx(0, 3)
amy_toffoli_qc.cx(0, 2)
amy_toffoli_qc.cx(2, 1)
amy_toffoli_qc.cx(1, 0)
amy_toffoli_qc.h(2)

# Implementation of the T-Depth 1 quasi-Toffoli found in https://arxiv.org/pdf/1210.0974.pdf
selinger_toffoli_qc = QuantumCircuit(4)
selinger_toffoli_qc.h(2)
selinger_toffoli_qc.cx(0, 3)
selinger_toffoli_qc.cx(2, [0, 1])
selinger_toffoli_qc.cx(1, 3)
selinger_toffoli_qc.t([0, 1])
selinger_toffoli_qc.t_dg([2, 3])
selinger_toffoli_qc.cx(1, 3)
selinger_toffoli_qc.cx(2, [0, 1])
selinger_toffoli_qc.cx(0, 3)
selinger_toffoli_qc.h(2)

# Implementation of the T-Depth 1 Toffoli found in https://arxiv.org/pdf/1212.5069.pdf
jones_toffoli_qc = QuantumCircuit(5, 1)
jones_toffoli_qc.append(
    selinger_toffoli_qc.to_gate("selinger_quasi_toffoli"), [0, 1, 3, 4]
)
jones_toffoli_qc.s(3)
jones_toffoli_qc.cx(3, 2)
jones_toffoli_qc.h(3)
jones_toffoli_qc.measure(3, 0)
jones_toffoli_qc.reset(3)
jones_toffoli_qc.append(CZGate().c_if(1), [0, 1], 0)



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from jax.lax import cond
import jax

from qrisp.environments import QuantumEnvironment
from qrisp.jasp import extract_invalues, insert_outvalues, check_for_tracing_mode


class ClControlEnvironment(QuantumEnvironment):
    r"""
    The ``ClControlEnvironment`` enables execution of quantum code conditioned on
    classical values. The environment works with similar semantics as the
    :ref:`ControlEnvironment`, implying this environment can also be entered
    using the ``control`` keyword.

    .. warning::

        Contrary to the :ref:`ControlEnvironment` the ``ClControlEnvironment`` must
        not have "carry values". This means that no value that is created inside this
        environment may be used outside of the environment.

    Examples
    ========

    We condition a quantum computation on the outcome of a previous measurement.

    ::

        from qrisp import *
        from qrisp.jasp import make_jaspr
        def test_f(i):

            a = QuantumFloat(3)
            a[:] = i
            b = measure(a)

            with control(b == 4):
                x(a[0])

            return measure(a)

        jaspr = make_jaspr(test_f)(1)

    This jaspr receives an integer and encodes that integer into the :ref:`QuantumFloat`
    `a`. Subsequently `a` is measured and an X gate is applied onto the 0-th
    qubit of `a` if the measurement value is 4.

    We can now evaluate the jaspr on several inputs

    >>> jaspr(1)
    1
    >>> jaspr(2)
    2
    >>> jaspr(3)
    3
    >>> jaspr(4)
    5

    We see that in the case where 4 was encoded, the X gate was indeed executed.

    To elaborate the restriction of carry values, we give an example that would
    be illegal:

    ::

        def test_f(i):

            a = QuantumFloat(3)
            a[:] = i
            b = measure(a)

            with control(b == 4):
                c = QuantumFloat(2)

            return measure(c)

        jaspr = make_jaspr(test_f)(1)

    This script creates a ``QuantumFloat`` `c` within the classical control
    environment and subsequently uses `c` outside of the environment (in the
    return statement).

    It is however possible to create (quantum-)values within the environment
    and use them still within the environment:

    ::

        from qrisp import *
        from qrisp.jasp import make_jaspr
        def test_f(i):

            a = QuantumFloat(3)
            a[:] = i
            b = measure(a)

            with control(b == 4):
                c = QuantumFloat(2)
                h(c[0])
                d = measure(c)

                # If c is measured to 1
                # flip a and uncompute c
                with control(d == 1):
                    x(a[0])
                    x(c[0])

                c.delete()

            return measure(a)

        jaspr = make_jaspr(test_f)(1)


    This script allocates another :ref:`QuantumFloat` `c` within the ClControlEnvironment
    and applies an Hadamard gate to the 0-th qubit. Subsequently the whole
    ``QuantumFloat`` is measured. If the measurement turns out to be one,
    the zeroth qubit of `a` is flipped (similar to the above examples) and
    furthermore `c` is brought back to the $\ket{0}$ state.

    >>> jaspr(4)
    5
    >>> jaspr(4)
    4




    """

    def __init__(self, ctrl_bls, ctrl_state=-1, invert=False):

        if not isinstance(ctrl_bls, list):
            ctrl_bls = [ctrl_bls]

        self.ctrl_bls = ctrl_bls

        QuantumEnvironment.__init__(self, env_args=ctrl_bls)

        # Process the ctrl_state
        self.ctrl_state = ctrl_state

        # If the ctrl state is a string, convert into an integer
        if isinstance(self.ctrl_state, str):
            if ctrl_state == len(ctrl_bls) * "1":
                self.ctrl_state = -1
            else:
                self.ctrl_state = int(self.ctrl_state, 2)

        self.ctrl_state = self.ctrl_state % (2 ** len(self.ctrl_bls))

        self.invert = invert

    def compile(self):
        for i in range(len(self.ctrl_bls)):
            if self.ctrl_bls[i] != bool((self.ctrl_state >> i) & 1):
                break
        else:
            QuantumEnvironment.compile(self)

    def __exit__(self, exception_type, exception_value, traceback):

        # "with" blocks in Python are always executed - the ClControlEnvironment
        # is no exception. It can therefore happen that code is run, which is not
        # intended to run by the semantics, because the classical condition is not
        # True. While this can not really prevented, it can lead to the
        # (annoying) behavior, that Exceptions are raised (for instance
        # out of bounds errors) from code that is not intended to be executed anyways.

        # If there was an exception raised, we mitigate this problem
        # by checking if the control condition was not True in the first place.
        # If this is the case, we simply exit the QuantumEnvironment without
        # any further action.
        static_error_appeared = False
        if not check_for_tracing_mode():
            if exception_type is not None:
                for i in range(len(self.ctrl_bls)):
                    ctrl_bl = self.ctrl_bls[i]
                    if (ctrl_bl ^ (self.ctrl_state >> i)) & 1:
                        self.env_qs.data = []
                        static_error_appeared = True
                        exception_type = None
                        exception_value = None
                        traceback = None
                        break

        QuantumEnvironment.__exit__(self, exception_type, exception_value, traceback)

        if static_error_appeared:
            return True

    def jcompile(self, eqn, context_dic):

        args = extract_invalues(eqn, context_dic)

        # This list stores the the variables representing the control variables
        ctrl_vars = args[: len(self.ctrl_bls)]

        # This list stores the variables used in the environment body
        env_vars = args[len(self.ctrl_bls) :]

        # Flatten the environments in the body
        body_jaspr = eqn.params["jaspr"].flatten_environments()

        if len(body_jaspr.outvars) > 1:
            raise Exception("Found ClControlEnvironment with carry value")

        # Compute the control bool
        tmp = ctrl_vars[0]
        cond_bl = tmp

        # Process the control state requirement
        if self.ctrl_state != -1 and ((self.ctrl_state & 1) == 0):
            cond_bl = ~tmp

        # If there is more than one control variable, loop through
        if len(ctrl_vars) > 1:

            for i in range(1, len(ctrl_vars)):
                tmp = ctrl_vars[i]
                if self.ctrl_state != -1 and ((self.ctrl_state & 1 << i) == 0):
                    tmp = ~tmp

                cond_bl = cond_bl & tmp

        if self.invert:
            cond_bl = ~cond_bl

        def identity_fun(*args):
            return args[-1]

        true_fun = identity_fun
        false_fun = identity_fun

        res_abs_qc = cond(cond_bl, true_fun, false_fun, *env_vars)

        insert_outvalues(eqn, context_dic, [res_abs_qc])

        traced_eqn = jax._src.core.thread_local_state.trace_state.trace_stack.dynamic.jaxpr_stack[
            0
        ].eqns[
            -1
        ]

        branch_0 = traced_eqn.params["branches"][0]
        branch_1 = traced_eqn.params["branches"][1]

        from qrisp.jasp import Jaspr

        traced_eqn.params["branches"] = (
            jax.core.ClosedJaxpr(Jaspr.from_cache(branch_0.jaxpr), branch_0.consts),
            jax.core.ClosedJaxpr(body_jaspr, branch_1.consts),
        )



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import jax

from qrisp.jasp.primitives import quantum_kernel_p, AbstractQubit, AbstractQubitArray
from qrisp.jasp.tracing_logic import TracingQuantumSession, qache


def quantum_kernel(func):
    """
    This decorator allows you to annotate a subroutine as a "quantum kernel".
    Quantum kernels are functions that are restricted in the sense that they
    can not have quantum inputs or outputs, yet their inner working can be quantum.
    What is the benefit in that? The underlying idea why this
    can be helpful is that future execution environments might host several
    QPUs that can operate in parallel, much like many of todays HPC
    environments can access multiple GPUs.

    Annotating a function as a quantum kernel therefore allows the execution
    environment to identify the subroutine as a separate quantum state and can
    assign it to a dedicated QPU.


    .. note::

        While not many quantum algorithms exist that directly allow such a
        parallelization, any sampling task can be performed in this manner:
        If you need to execute a 1000 shots of a certain quantum circuit, but
        you have 4 QPUs available, you can execute the task 4 times faster by
        assigning 250 shots to each QPU.
        As such the :ref:`sample <sample>` and :ref:`expectation_value <expectation_value>` function
        automatically wraps the state preparation and measurement into a
        dedicated quantum kernel.


    Parameters
    ----------
    func : callable
        A function that receives only classical values as inputs and returns
        classical values as outpus. The function's body can however perform
        quantum logic.

    Returns
    -------
    quantum_kernel : callable
        A function that performs the task of the input but the compiler can
        identify it as a closed quantum procedure without any external entaglement.

    Examples
    --------

    We demonstrate a naive implementation of an expectation value please use
    :ref:`expectation_value` if you required this functionality. For this
    we define a state preparation procedure and call it from a kernelized
    function.

    ::

        from qrisp import *
        from qrisp.jasp import *

        def state_prep(k):
            qf = QuantumFloat(5)
            h(qf[k])
            return qf

        @quantum_kernel
        def sampling_kernel(k):
            # Receives a classical (!) integer k

            qf = state_prep(k)

            # Returns a classical integer
            return measure(qf)

    We now call the kernel within a purely classical Jax script.

    ::

        @jaspify
        def main(k):

            shots = 100

            res = 0
            for i in range(shots):
                res += sampling_kernel(k)

            return res/shots

    Perform some experiments:

    ::

        print(main(3))
        # Yields: 3.92
        # Expected: 2**3/2 = 4
        print(main(4))
        # Yields: 8.96
        # Expected: 2**4/2 = 8

    """

    func = qache(func)

    def return_function(*args, **kwargs):

        from qrisp.jasp.jasp_expression.centerclass import Jaspr, collect_environments

        qs = TracingQuantumSession.get_instance()

        qs.start_tracing(quantum_kernel_p.bind(), gc_mode = "none")

        try:
            res = func(*args, **kwargs)
        except Exception as e:
            qs.conclude_tracing()
            raise e

        eqn = jax._src.core.thread_local_state.trace_state.trace_stack.dynamic.jaxpr_stack[
            0
        ].eqns[
            -1
        ]
        flattened_jaspr = Jaspr.from_cache(
            collect_environments(eqn.params["jaxpr"].jaxpr)
        ).flatten_environments()
        for var in flattened_jaspr.invars:
            if isinstance(var.aval, (AbstractQubitArray, AbstractQubit)):
                raise Exception("Tried to construct quantum kernel with quantum input")
        for var in flattened_jaspr.outvars:
            if isinstance(var.aval, (AbstractQubitArray, AbstractQubit)):
                raise Exception("Tried to construct quantum kernel with quantum output")

        eqn.params["jaxpr"] = jax.core.ClosedJaxpr(
            flattened_jaspr, eqn.params["jaxpr"].consts
        )

        qs.conclude_tracing()

        return res

    return return_function



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import inspect

import jax
import jax.numpy as jnp

from qrisp.environments.quantum_environments import QuantumEnvironment
from qrisp.environments.gate_wrap_environment import GateWrapEnvironment
from qrisp.circuit import Operation, QuantumCircuit, Instruction
from qrisp.environments.iteration_environment import IterationEnvironment
from qrisp.core import merge

from qrisp.jasp import check_for_tracing_mode, qache, AbstractQubit, make_jaspr


def custom_inversion(*func, **cusi_kwargs):
    """
    The ``custom_control`` decorator allows to specify the controlled version of
    the decorated function. If this function is called within a :ref:`ControlEnvironment`
    or a :ref:`ConditionEnvironment` the controlled version is executed instead.

    Specific controlled versions of quantum functions are very common in many
    scientific publications. This is because the general control procedure can
    signifcantly increase resource demands.

    In order to use the ``custom_control`` decorator, you need to add the ``ctrl``
    keyword to your function signature. If called within a controlled context,
    this keyword will receive the corresponding control qubit.

    For more details consult the examples section.


    Parameters
    ----------
    func : function
        A function of QuantumVariables, which has the ``ctrl`` keyword.

    Returns
    -------
    adaptive_control_function : function
        A function which will execute it's controlled version, if called
        within a :ref:`ControlEnvironment` or a :ref:`ConditionEnvironment`.

    Examples
    --------

    We create a swap function with custom control.

    ::

        from qrisp import mcx, cx, custom_control

        @custom_control
        def swap(a, b, ctrl = None):

            if ctrl is None:

                cx(a, b)
                cx(b, a)
                cx(a, b)

            else:

                cx(a, b)
                mcx([ctrl, b], a)
                cx(a, b)


    Test the non-controlled version:

    ::

        from qrisp import QuantumBool

        a = QuantumBool()
        b = QuantumBool()

        swap(a, b)

        print(a.qs)


    ::

        QuantumCircuit:
        --------------
                  ┌───┐
        a.0: ──■──┤ X ├──■──
             ┌─┴─┐└─┬─┘┌─┴─┐
        b.0: ┤ X ├──■──┤ X ├
             └───┘     └───┘
        Live QuantumVariables:
        ---------------------
        QuantumBool a
        QuantumBool b


    Test the controlled version:

    ::

        from qrisp import control

        a = QuantumBool()
        b = QuantumBool()
        ctrl_qbl = QuantumBool()

        with control(ctrl_qbl):

            swap(a,b)

        print(a.qs.transpile(1))

    ::

                         ┌───┐
               a.0: ──■──┤ X ├──■──
                    ┌─┴─┐└─┬─┘┌─┴─┐
               b.0: ┤ X ├──■──┤ X ├
                    └───┘  │  └───┘
        ctrl_qbl.0: ───────■───────


    """

    if len(func) == 0:
        return lambda x: custom_inversion(x, **cusi_kwargs)
    else:
        func = func[0]

    # The idea to realize the custom control feature in traced mode is to
    # first trace the non-controlled version into a pjit primitive using
    # the qache feature and the trace the controlled version.
    # The controlled version is then stored in the params attribute

    # Qache the function (in non-traced mode, this has no effect)
    
    # Make sure the inv keyword argument is treated as a static argument
    new_static_argnames = list(cusi_kwargs.get("static_argnames", []))
    new_static_argnames.append("inv")
    cusi_kwargs["static_argnames"] = new_static_argnames
    
    qached_func = qache(func, **cusi_kwargs)

    def adaptive_inversion_function(*args, **kwargs):

        if not check_for_tracing_mode():
            return func(*args, **kwargs)

        else:

            args = list(args)
            for i in range(len(args)):
                if isinstance(args[i], bool):
                    args[i] = jnp.array(args[i], dtype=jnp.bool)
                elif isinstance(args[i], int):
                    args[i] = jnp.array(args[i], dtype=jnp.int64)
                elif isinstance(args[i], float):
                    args[i] = jnp.array(args[i], dtype=jnp.float64)
                elif isinstance(args[i], complex):
                    args[i] = jnp.array(args[i], dtype=jnp.complex64)

            # Call the (qached) function
            res = qached_func(*args, inv = False, **kwargs)

            # Retrieve the pjit equation
            jit_eqn = jax._src.core.thread_local_state.trace_state.trace_stack.dynamic.jaxpr_stack[
                0
            ].eqns[
                -1
            ]

            if not jit_eqn.params["jaxpr"].jaxpr.inv_jaspr:
                # Trace the inverted version

                def ammended_func(*args, **kwargs):
                    new_kwargs = dict(kwargs)
                    return func(*args, inv = True, **new_kwargs)

                inverted_jaspr = make_jaspr(ammended_func)(
                    *args, **kwargs
                )
                
                # Store controlled version
                jit_eqn.params["jaxpr"].jaxpr.inv_jaspr = inverted_jaspr
                inverted_jaspr.inv_jaspr = jit_eqn.params["jaxpr"].jaxpr

        return res

    return adaptive_inversion_function

